__all__ = [
    'NumberFieldHandler',
    'StringFieldHandler',
    'URLFieldHandler',
    'EmailFieldHandler',
    'DecimalFieldHandler',
    'ComplexDateTimeFieldHandler',
    'ListFieldHandler',
    'DictFieldHandler',
    'BinaryFieldHandler',
    'SequenceFieldHandler',
    'UUIDFieldHandler',
    'ReferenceFieldHandler',
    'CachedReferenceFieldHandler',
    'FileFieldHandler',
    'ImageFieldHandler',
    'EmbeddedDocumentFieldHandler',
    'EmbeddedDocumentListFieldHandler'
]

import re
from typing import Type, Collection, Union

import bson
import mongoengine.fields

from mongoengine_migrate.exceptions import SchemaError, InconsistencyError
from mongoengine_migrate.mongo import (
    check_empty_result,
    mongo_version
)
from ..updater import ByPathContext, ByDocContext, DocumentUpdater
from mongoengine_migrate.utils import get_document_type, Diff, UNSET
from .base import CommonFieldHandler
from .converters import to_string, to_decimal


class NumberFieldHandler(CommonFieldHandler):
    field_classes = [
        mongoengine.fields.IntField,
        mongoengine.fields.LongField,
        mongoengine.fields.FloatField,
    ]
    schema_skel_keys = {'min_value', 'max_value'}

    def change_min_value(self, updater: DocumentUpdater, diff: Diff):
        """
        Change min_value of field. Force set to minimum if value is
        less than limitation (if any)
        """
        def by_path(ctx: ByPathContext):
            ctx.collection.update_many(
                {ctx.filter_dotpath: {'$lt': diff.new}, **ctx.extra_filter},
                {'$set': {ctx.update_dotpath: diff.new}},
                array_filters=ctx.build_array_filters()
            )

        self._check_diff(updater, diff, True, (int, float))
        if diff.new in (UNSET, None):
            return

        updater.update_by_path(by_path)

    def change_max_value(self, updater: DocumentUpdater, diff: Diff):
        """
        Change max_value of field. Force set to maximum if value is
        more than limitation (if any)
        """
        def by_path(ctx: ByPathContext):
            ctx.collection.update_many(
                {ctx.filter_dotpath: {'$gt': diff.new}, **ctx.extra_filter},
                {'$set': {ctx.update_dotpath: diff.new}},
                array_filters=ctx.build_array_filters()
            )

        self._check_diff(updater, diff, True, (int, float))
        if diff.new in (UNSET, None):
            return

        updater.update_by_path(by_path)


class StringFieldHandler(CommonFieldHandler):
    field_classes = [
        mongoengine.fields.StringField,
    ]

    schema_skel_keys = {'max_length', 'min_length', 'regex'}

    def change_max_length(self, updater: DocumentUpdater, diff: Diff):
        """Cut off a string if it longer than limitation (if any)"""
        def by_doc(ctx: ByDocContext):
            doc = ctx.document
            match = updater.field_name in doc and len(doc[updater.field_name]) > diff.new
            if match:
                doc[updater.field_name] = doc[updater.field_name][:diff.new]

        self._check_diff(updater, diff, True, int)
        if diff.new in (UNSET, None):
            return
        if diff.new < 0:
            diff.new = 0

        # Cut too long strings
        updater.update_by_document(by_doc)

    @mongo_version(min_version='3.6')
    def change_min_length(self, updater: DocumentUpdater, diff: Diff):
        """Raise error if string is shorter than limitation (if any)"""
        def by_path(ctx: ByPathContext):
            fltr = {
                ctx.filter_dotpath: {'$ne': None},
                **ctx.extra_filter,
                "$expr": {"$lt": [{"$strLenCP": f"${ctx.filter_dotpath}"}, diff.new]},  # >= 3.6
            }
            check_empty_result(ctx.collection, ctx.filter_dotpath, fltr)

        def by_doc(ctx: ByDocContext):
            doc = ctx.document
            val = doc.get(updater.field_name)
            if isinstance(val, str) and len(val) < diff.new:
                raise InconsistencyError(
                    f"String field {ctx.collection.name}.{ctx.filter_dotpath} on one of "
                    f"records has length less than minimum: {doc}"
                )

        self._check_diff(updater, diff, True, int)
        if diff.new in (UNSET, None):
            return
        if diff.new < 0:
            diff.new = 0

        # We can't to increase string length, so raise error if
        # there was found strings which are shorter than should be
        if self.migration_policy.name == 'strict':
            updater.update_combined(by_path, by_doc, False)

    def change_regex(self, updater: DocumentUpdater, diff: Diff):
        """Raise error if string does not match regex (if any)"""
        def by_path(ctx: ByPathContext):
            fltr = {
                ctx.filter_dotpath: {'$not': re.compile(diff.new), '$ne': None, '$exists': True},
                **ctx.extra_filter
            }
            check_empty_result(ctx.collection, ctx.filter_dotpath, fltr)

        self._check_diff(updater, diff, True, (str, type(re.compile('.'))))
        if diff.new in (UNSET, None):
            return

        if self.migration_policy.name == 'strict':
            updater.update_by_path(by_path)


class URLFieldHandler(StringFieldHandler):
    field_classes = [
        mongoengine.fields.URLField,
    ]

    schema_skel_keys = {'schemes'}  # TODO: implement url_regex

    def change_schemes(self, updater: DocumentUpdater, diff: Diff):
        """Raise error if url has scheme not from list"""
        def by_path(ctx: ByPathContext):
            fltr = {ctx.filter_dotpath: {'$not': scheme_regex, '$ne': None}, **ctx.extra_filter}
            check_empty_result(ctx.collection, ctx.filter_dotpath, fltr)

        self._check_diff(updater, diff, False, Collection)
        if not diff.new or diff.new == UNSET:
            return

        # Check if some records contains non-url values in db_field
        scheme_regex = re.compile(rf'\A(?:({"|".join(re.escape(x) for x in diff.new)}))://')
        if self.migration_policy.name == 'strict':
            updater.update_by_path(by_path)


class EmailFieldHandler(StringFieldHandler):
    field_classes = [
        mongoengine.fields.EmailField
    ]

    schema_skel_keys = {'domain_whitelist', 'allow_utf8_user', 'allow_ip_domain'}

    USER_REGEX = re.compile(
        # `dot-atom` defined in RFC 5322 Section 3.2.3.
        r"(\A[-!#$%&'*+/=?^_`{}|~0-9A-Z]+(\.[-!#$%&'*+/=?^_`{}|~0-9A-Z]+)*@.+\Z"
        # `quoted-string` defined in RFC 5322 Section 3.2.4.
        r'|\A"([\001-\010\013\014\016-\037!#-\[\]-\177]|\\[\001-\011\013\014\016-\177])*"@.+\Z)',
        re.IGNORECASE,
    )

    UTF8_USER_REGEX = re.compile(
        # RFC 6531 Section 3.3 extends `atext` (used by dot-atom) to
        # include `UTF8-non-ascii`.
        r"(\A[-!#$%&'*+/=?^_`{}|~0-9A-Z\u0080-\U0010FFFF]+(\.[-!#$%&'*+/=?^_`{}|~0-9A-Z\u0080-\U0010FFFF]+)*@.+\Z"
        # `quoted-string`
        r'|\A"([\001-\010\013\014\016-\037!#-\[\]-\177]|\\[\001-\011\013\014\016-\177])*"@.+\Z)',
        re.IGNORECASE | re.UNICODE,
    )

    IP_DOMAIN_REGEX = re.compile(
        # ipv6
        r'\A[^@]+@\[(::)?([A-F0-9]{1,4}::?){0,7}([A-F0-9]{1,4})?\]\Z'
        # ipv4
        r'|\A[^@]+@\[\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}\]\Z'
    )

    DOMAIN_REGEX = re.compile(
        r"\A[^@]+@((?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+)(?:[A-Z0-9-]{2,63}(?<!-))\Z",
        re.IGNORECASE,
    )

    def change_domain_whitelist(self, updater: DocumentUpdater, diff: Diff):
        """
        `domain_whitelist` parameter is affected to domain validation:
        if domain in email address is in that list then validation will
        be skipped. So, do nothing here
        """

    def change_allow_utf8_user(self, updater: DocumentUpdater, diff: Diff):
        """Raise error if email address has wrong user name"""
        def by_path(ctx: ByPathContext):
            # Find records which doesn't match to the regex
            fltr = {ctx.filter_dotpath: {'$not': regex, '$ne': None}, **ctx.extra_filter}
            check_empty_result(ctx.collection, ctx.filter_dotpath, fltr)

        self._check_diff(updater, diff, False, bool)
        if diff.new == UNSET:
            return

        regex = self.UTF8_USER_REGEX if diff.new is True else self.USER_REGEX
        if self.migration_policy.name == 'strict':
            updater.update_by_path(by_path)

    def change_allow_ip_domain(self, updater: DocumentUpdater, diff: Diff):
        """
        Raise error if email has domain which not in `domain_whitelist`
        when `allow_ip_domain` is True. Otherwise do nothing
        """
        def by_path(ctx: ByPathContext):
            # Find records with ip domains and raise error if found
            fltr = {"$and": [
                {ctx.filter_dotpath: {'$ne': None}},
                {ctx.filter_dotpath: self.IP_DOMAIN_REGEX},
                {ctx.filter_dotpath: {'$not': re.compile(rf'\A[^@]+@({whitelist_regex})\Z')}},
                *[{k: v} for k, v in ctx.extra_filter.items()]
            ]}
            check_empty_result(ctx.collection, ctx.filter_dotpath, fltr)

        self._check_diff(updater, diff, False, bool)
        if diff.new is True or diff.new == UNSET:
            return

        whitelist_regex = '|'.join(
            re.escape(x) for x in self.left_field_schema.get('domain_whitelist', [])
        ) or '.*'
        if self.migration_policy.name == 'strict':
            updater.update_by_path(by_path)

    def convert_type(self,
                     updater: DocumentUpdater,
                     from_field_cls: Type[mongoengine.fields.BaseField],
                     to_field_cls: Type[mongoengine.fields.BaseField]):
        def by_path(ctx: ByPathContext):
            fltr = {'$and': [
                {ctx.filter_dotpath: {'$ne': None}},
                {ctx.filter_dotpath: {'$not': self.DOMAIN_REGEX}},
                {ctx.filter_dotpath: {'$not': self.IP_DOMAIN_REGEX}},
                {ctx.filter_dotpath: {'$not': re.compile(rf'\A[^@]+@({whitelist_regex})\Z')}},
                *[{k: v} for k, v in ctx.extra_filter.items()],
            ]}
            check_empty_result(ctx.collection, ctx.filter_dotpath, fltr)

        to_string(updater)

        # Find records with ip domains and raise error if found
        whitelist_regex = '|'.join(
            re.escape(x) for x in self.left_field_schema.get('domain_whitelist', [])
        ) or '.*'
        if self.migration_policy.name == 'strict':
            updater.update_by_path(by_path)


class DecimalFieldHandler(NumberFieldHandler):
    field_classes = [mongoengine.fields.DecimalField]

    schema_skel_keys = {'force_string', 'precision', 'rounding'}

    def change_force_string(self, updater: DocumentUpdater, diff: Diff):
        """
        Convert to string or decimal depending on `force_string` flag
        """
        self._check_diff(updater, diff, False, bool)
        if diff.new == UNSET:
            return

        if diff.new is True:
            to_string(updater)
        else:
            to_decimal(updater)

    def change_precision(self, updater: DocumentUpdater, diff: Diff):
        """This one is related only for python. Nothing to do"""
        pass

    def change_rounding(self, updater: DocumentUpdater, diff: Diff):
        """This one is related only for python. Nothing to do"""
        pass

    def convert_type(self,
                     updater: DocumentUpdater,
                     from_field_cls: Type[mongoengine.fields.BaseField],
                     to_field_cls: Type[mongoengine.fields.BaseField]):
        if self.left_field_schema.get('force_string', True):
            to_string(updater)
        else:
            to_decimal(updater)


class ComplexDateTimeFieldHandler(StringFieldHandler):
    field_classes = [mongoengine.fields.ComplexDateTimeField]

    schema_skel_keys = {'separator'}

    def change_separator(self, updater: DocumentUpdater, diff: Diff):
        """Change separator in datetime strings"""
        def by_doc(ctx: ByDocContext):
            doc = ctx.document
            if updater.field_name in doc:
                doc[updater.field_name] = doc[updater.field_name].replace(diff.old, diff.new)

        self._check_diff(updater, diff, False, str)
        if not diff.new or not diff.old:
            raise SchemaError(f'{updater.document_type}{updater.field_name}.separator '
                              f'must not be empty')
        if diff.new == UNSET:
            return

        updater.update_by_document(by_doc)


class ListFieldHandler(CommonFieldHandler):
    field_classes = [
        mongoengine.fields.ListField
    ]

    schema_skel_keys = set()  # TODO: implement "field"

    @classmethod
    def schema_skel(cls) -> dict:
        """
        Return db schema skeleton dict, which contains keys taken from
        `schema_skel_keys` and Nones as values
        """
        skel = super(ListFieldHandler, cls).schema_skel()

        # Add optional keys to skel if they are set in field object
        # * `max_length` was added in mongoengine 0.19.0
        optional_keys = {'max_length'}
        assert all(issubclass(c, mongoengine.fields.ListField) for c in cls.field_classes), \
            'If you wanna add field class then this code should be rewritten'
        field_obj = mongoengine.fields.ListField()
        skel.update({k: None for k in optional_keys if hasattr(field_obj, k)})

        return skel

    def change_max_length(self, updater: DocumentUpdater, diff: Diff):
        """Cut off a list if it longer than limitation (if any)"""
        def by_doc(ctx: ByDocContext):
            doc = ctx.document
            match = updater.field_name in doc and len(doc[updater.field_name]) > diff.new
            if match:
                doc[updater.field_name] = doc[updater.field_name][:diff.new]

        self._check_diff(updater, diff, True, int)
        if diff.new in (UNSET, None):
            return

        updater.update_by_document(by_doc)


class DictFieldHandler(CommonFieldHandler):
    pass  # TODO: implement "field" param


class BinaryFieldHandler(CommonFieldHandler):
    field_classes = [
        mongoengine.fields.BinaryField
    ]

    schema_skel_keys = {'max_bytes'}

    def change_max_bytes(self, updater: DocumentUpdater, diff: Diff):
        """
        $binarySize expression is not available in MongoDB yet,
        so do nothing
        """
        # TODO: add python update
        self._check_diff(updater, diff, True, int)
        if diff.new in (UNSET, None):
            return

        pass


class SequenceFieldHandler(CommonFieldHandler):
    field_classes = [
        mongoengine.fields.SequenceField
    ]

    # TODO: warning on using non-default value_decorator
    # TODO: db_alias
    schema_skel_keys = {'collection_name', 'sequence_name'}

    def change_collection_name(self, updater: DocumentUpdater, diff: Diff):
        """Typically changing the collection name should not require
        to do any changes
        """
        self._check_diff(updater, diff, False, str)
        pass

    def change_sequence_name(self, updater: DocumentUpdater, diff: Diff):
        """Typically changing the sequence name should not require
        to do any changes
        """
        self._check_diff(updater, diff, False, str)
        pass


class UUIDFieldHandler(CommonFieldHandler):
    field_classes = [
        mongoengine.fields.UUIDField
    ]

    schema_skel_keys = {'binary'}

    def change_binary(self, updater: DocumentUpdater, diff: Diff):
        self._check_diff(updater, diff, False, bool)

        if diff.new is True:
            pass
            # TODO: convert to bson Binary
        else:
            pass
            # TODO: convert Binary to string


class ReferenceFieldHandler(CommonFieldHandler):
    field_classes = [
        mongoengine.fields.ReferenceField,
        mongoengine.fields.LazyReferenceField
    ]

    schema_skel_keys = {'target_doctype', 'dbref'}

    def change_target_doctype(self, updater: DocumentUpdater, diff: Diff):
        """Collection could not exist in db, so do nothing"""
        self._check_diff(updater, diff, False, str)

    def change_dbref(self, updater: DocumentUpdater, diff: Diff):
        """Change reference storing format: ObjectId or DBRef"""
        self._check_diff(updater, diff, False, bool)

        if diff.new is True:
            self._objectid_to_dbref(updater)
        else:
            self._dbref_to_objectid(updater)

    def _objectid_to_dbref(self, updater: DocumentUpdater):
        def by_doc(ctx: ByDocContext):
            doc = ctx.document
            if isinstance(doc.get(updater.field_name), bson.ObjectId):
                doc[updater.field_name] = bson.DBRef(ctx.collection.name, doc[updater.field_name])

        updater.update_by_document(by_doc)

    def _dbref_to_objectid(self, updater: DocumentUpdater):
        def by_doc(ctx: ByDocContext):
            doc = ctx.document
            if isinstance(doc.get(updater.field_name), bson.DBRef):
                doc[updater.field_name] = doc[updater.field_name].id

        updater.update_by_document(by_doc)

    @classmethod
    def build_schema(
            cls,
            field_obj: Union[
                mongoengine.fields.ReferenceField,
                mongoengine.fields.LazyReferenceField
            ]) -> dict:
        schema = super(ReferenceFieldHandler, cls).build_schema(field_obj)

        # 'document_type' is restricted to use Document class
        # as value by mongoengine itself
        document_type_cls = field_obj.document_type
        schema['target_doctype'] = get_document_type(document_type_cls)

        return schema


class CachedReferenceFieldHandler(CommonFieldHandler):
    field_classes = [
        mongoengine.fields.CachedReferenceField
    ]

    schema_skel_keys = {'target_doctype', 'fields'}

    def change_target_doctype(self, updater: DocumentUpdater, diff: Diff):
        """Collection could not exist in db, so do nothing"""
        self._check_diff(updater, diff, False, str)

    def change_fields(self, updater: DocumentUpdater, diff: Diff):
        def by_doc(ctx: ByDocContext):
            if updater.field_name in ctx.document:
                if isinstance(ctx.document[updater.field_name], dict):
                    ctx.document[updater.field_name] = {
                        k: v for k, v in ctx.document[updater.field_name].items()
                        if k in keep_fields
                    }

        self._check_diff(updater, diff, False, (list, tuple))

        if diff.new:
            keep_fields = set(diff.new)
            if not updater.is_embedded:
                keep_fields.add('_id')
            updater.update_by_document(by_doc)

    @classmethod
    def build_schema(
            cls,
            field_obj: Union[
                mongoengine.fields.ReferenceField,
                mongoengine.fields.LazyReferenceField
            ]) -> dict:
        schema = super(CachedReferenceFieldHandler, cls).build_schema(field_obj)

        # 'document_type' is restricted to use Document class
        # as value by mongoengine itself
        document_type_cls = field_obj.document_type
        schema['target_doctype'] = get_document_type(document_type_cls)

        return schema


class FileFieldHandler(CommonFieldHandler):
    field_classes = [
        mongoengine.fields.FileField
    ]

    # TODO: db_alias
    schema_skel_keys = {'collection_name'}

    def change_collection_name(self, updater: DocumentUpdater, diff: Diff):
        """Typically changing the collection name should not require
        to do any changes
        """
        self._check_diff(updater, diff, False, str)
        pass


class ImageFieldHandler(FileFieldHandler):
    field_classes = [
        mongoengine.fields.ImageField
    ]

    schema_skel_keys = {'size', 'thumbnail_size'}

    def change_thumbnail_size(self, updater: DocumentUpdater, diff: Diff):
        """Typically changing the attribute should not require
        to do any changes
        """
        self._check_diff(updater, diff, False, (list, tuple))
        pass

    def change_size(self, updater: DocumentUpdater, diff: Diff):
        """Typically changing the attribute should not require
        to do any changes
        """
        self._check_diff(updater, diff, False, (list, tuple))
        pass


class EmbeddedDocumentFieldHandler(CommonFieldHandler):
    field_classes = [
        mongoengine.fields.EmbeddedDocumentField
    ]

    schema_skel_keys = {'target_doctype'}

    def change_target_doctype(self, updater: DocumentUpdater, diff: Diff):
        self._check_diff(updater, diff, False, str)
        # FIXME: decide what to do with existed embedded docs?

    @classmethod
    def build_schema(cls, field_obj: mongoengine.fields.EmbeddedDocumentField) -> dict:
        schema = super(EmbeddedDocumentFieldHandler, cls).build_schema(field_obj)

        document_type_cls = field_obj.document_type
        schema['target_doctype'] = get_document_type(document_type_cls)

        return schema


class EmbeddedDocumentListFieldHandler(ListFieldHandler):
    field_classes = [
        mongoengine.fields.EmbeddedDocumentListField
    ]

    # FIXME: move 'target_doctype' changing from here to ListField 'db_field'
    #        this require to fix embedded document recursive walk method
    #        to also consider field.document_type of ListField
    #        This will help to manage with situation when ListField
    #        with EmbeddedDocumentField is defined by user
    schema_skel_keys = {'target_doctype'}

    def change_target_doctype(self, updater: DocumentUpdater, diff: Diff):
        self._check_diff(updater, diff, False, str)
        # FIXME: decide what to do with existed embedded docs?

    @classmethod
    def build_schema(cls, field_obj: mongoengine.fields.EmbeddedDocumentListField) -> dict:
        schema = super(EmbeddedDocumentListFieldHandler, cls).build_schema(field_obj)

        document_type_cls = field_obj.field.document_type
        schema['target_doctype'] = get_document_type(document_type_cls)

        return schema



__all__ = [
    'CreateDocument',
    'DropDocument',
    'RenameDocument',
    'AlterDocument'
]

import logging

from mongoengine_migrate.flags import EMBEDDED_DOCUMENT_NAME_PREFIX
from mongoengine_migrate.schema import Schema
from mongoengine_migrate.updater import DocumentUpdater, ByDocContext, ByPathContext
from mongoengine_migrate.utils import Diff
from .base import BaseCreateDocument, BaseDropDocument, BaseRenameDocument, BaseAlterDocument

log = logging.getLogger('mongoengine-migrate')


class CreateDocument(BaseCreateDocument):
    """Create new document in db
    # FIXME: parameters (indexes, acl, etc.)
    """
    priority = 8

    @classmethod
    def build_object(cls, document_type: str, left_schema: Schema, right_schema: Schema):
        if document_type.startswith(EMBEDDED_DOCUMENT_NAME_PREFIX):
            # This is an embedded document
            return None

        return super().build_object(document_type, left_schema, right_schema)

    def run_forward(self):
        """
        Mongodb automatically creates collection on the first insert
        So, do nothing
        FIXME: parameters (indexes, acl, etc.)
        """

    def run_backward(self):
        """Drop collection in backward direction"""
        # If the document has 'allow_inheritance' then drop only if
        # no derived or parent documents left which are point to
        # the same collection
        skip = self.parameters.get('inherit') and self._is_my_collection_used_by_other_documents()
        if not skip:
            self._run_ctx['collection'].drop()
        # FIXME: add removing by _cls


class DropDocument(BaseDropDocument):
    """Drop a document"""
    priority = 16

    @classmethod
    def build_object(cls, document_type: str, left_schema: Schema, right_schema: Schema):
        if document_type.startswith(EMBEDDED_DOCUMENT_NAME_PREFIX):
            # This is an embedded document
            return None

        return super(DropDocument, cls).build_object(document_type, left_schema, right_schema)

    def run_forward(self):
        """
        Mongodb automatically creates collection on the first insert
        So, do nothing
        """
        # If the document has 'allow_inheritance' then drop only if
        # no derived or parent documents left which are point to
        # the same collection
        skip = self.parameters.get('inherit') and self._is_my_collection_used_by_other_documents()
        if not skip:
            self._run_ctx['collection'].drop()
        # FIXME: add removing by _cls

    def run_backward(self):
        """
        Mongodb automatically creates collection on the first insert
        So, do nothing
        FIXME: parameters (indexes, acl, etc.)
        """


class RenameDocument(BaseRenameDocument):
    """Rename document"""
    priority = 6

    @classmethod
    def build_object(cls, document_type: str, left_schema: Schema, right_schema: Schema):
        if document_type.startswith(EMBEDDED_DOCUMENT_NAME_PREFIX):
            # This is an embedded document
            return None

        return super(RenameDocument, cls).build_object(document_type, left_schema, right_schema)

    def run_forward(self):
        """Rename document only in schema, so do nothing"""

    def run_backward(self):
        """Rename document only in schema, so do nothing"""


class AlterDocument(BaseAlterDocument):
    priority = 9

    @classmethod
    def build_object(cls, document_type: str, left_schema: Schema, right_schema: Schema):
        if document_type.startswith(EMBEDDED_DOCUMENT_NAME_PREFIX):
            # This is an embedded document
            return None

        return super(AlterDocument, cls).build_object(document_type, left_schema, right_schema)

    def change_collection(self, updater: DocumentUpdater, diff: Diff):
        self._check_diff(diff, False, str)

        old_collection = self._run_ctx['db'][diff.old]
        collection_names = self._run_ctx['collection'].database.list_collection_names()

        # If the document has 'allow_inheritance' then rename only if
        # no documents left which are point to collection
        skip = self.parameters.get('inherit') and self._is_my_collection_used_by_other_documents()
        if not skip and diff.old in collection_names:
            old_collection.rename(diff.new)
            # Update collection object in run context after renaming
            self._run_ctx['collection'] = self._run_ctx['db'][diff.new]

    def change_inherit(self, updater: DocumentUpdater, diff: Diff):
        """Remove '_cls' key if Document becomes non-inherit, otherwise
        do nothing
        """
        def by_path(ctx: ByPathContext):
            ctx.collection.update_many(
                {ctx.filter_dotpath + '._cls': {'$exists': True}, **ctx.extra_filter},
                {'$unset': {ctx.update_dotpath + '._cls': ''}},
                array_filters=ctx.build_array_filters()
            )

        self._check_diff(diff, False, bool)
        if diff.new:
            return

        updater.update_by_path(by_path)

    def change_dynamic(self, updater: DocumentUpdater, diff: Diff):
        """If document becomes non-dynamic then remove fields which
        are not defined in mongoengine Document
        """
        def by_doc(ctx: ByDocContext):
            extra_keys = ctx.document.keys() - self_schema.keys()
            if extra_keys:
                newdoc = {k: v for k, v in ctx.document.items() if k in self_schema.keys()}
                ctx.document.clear()
                ctx.document.update(newdoc)

        self._check_diff(diff, False, bool)
        if diff.new:
            return  # Nothing to do

        # Remove fields which are not in schema
        self_schema = self._run_ctx['left_schema'][self.document_type]  # type: Schema.Document
        updater.update_by_document(by_doc)



__all__ = [
    'RunPython'
]

import logging

from mongoengine_migrate.exceptions import ActionError
from mongoengine_migrate.schema import Schema
from .base import BaseAction

log = logging.getLogger('mongoengine-migrate')


class RunPython(BaseAction):
    """
    Action which runs user defined functions. It's supposed
    that user must also handle possible schema changes in his functions.
    """
    # TODO: implement handling of user's schema changes and to_schema_patch
    def __init__(self, document_type, *, forward_func=None, backward_func=None, **kwargs):
        super().__init__(document_type, **kwargs)

        if forward_func is None and backward_func is None:
            raise ActionError("forward_func and backward_func are not set")

        self.forward_func = forward_func
        self.backward_func = backward_func

    def run_forward(self):
        if self.forward_func is not None:
            self.forward_func(self._run_ctx['db'],
                              self._run_ctx['collection'],
                              self._run_ctx['left_schema'])

    def run_backward(self):
        if self.backward_func is not None:
            self.backward_func(self._run_ctx['db'],
                               self._run_ctx['collection'],
                               self._run_ctx['left_schema'])

    def to_schema_patch(self, left_schema: Schema):
        """
        We can't predict what code will be placed to user functions
        so don't assume any schema changes
        """
        return []

    def to_python_expr(self) -> str:
        parameters = {
            name: getattr(val, 'to_python_expr', lambda: repr(val))()
            for name, val in self.parameters.items()
        }
        kwargs_str = ''.join(f", {name}={val}" for name, val in sorted(parameters.items()))
        ff_expr = f'forward_func={self.forward_func.__name__ if self.forward_func else None}'
        bf_expr = f'backward_func={self.backward_func.__name__ if self.backward_func else None}'
        return f'{self.__class__.__name__}({self.document_type!r}, ' \
               f'{ff_expr + ", " if self.forward_func else ""}' \
               f'{bf_expr + ", " if self.backward_func else ""}' \
               f'{kwargs_str})'



#!/usr/bin/env python3
import logging
import sys
from typing import Optional

import click

import mongoengine_migrate.flags as flags
from mongoengine_migrate.loader import MongoengineMigrate, import_module
from mongoengine_migrate.exceptions import MongoengineMigrateError

mongoengine_migrate: Optional[MongoengineMigrate] = None


log = logging.getLogger('mongoengine-migrate')


def setup_logger(log_level: str):
    fmt = logging.Formatter('[%(levelname)s] %(message)s')
    hdlr = logging.StreamHandler()
    hdlr.setFormatter(fmt)
    log.addHandler(hdlr)

    log.setLevel(log_level.upper())


def error_handler(exception: MongoengineMigrateError) -> None:
    """
    MongoengineMigrateError and derived errors handler. Depending on
    current log level either reraise those exception or print error
    to logger and exit with error code
    :param exception:
    :return:
    """
    if log.level == logging.DEBUG:
        raise exception

    log.error('The following error has occured during executing (use `--log-level=debug` '
              'argument to get more info): %s: %s',
              exception.__class__.__name__,
              str(exception))
    sys.exit(1)


def cli_options(f):
    decorators = [
        click.option(
            "-u",
            "--uri",
            default='mongodb://localhost/mydb',
            envvar="MONGOENGINE_MIGRATE_URI",
            metavar="URI",
            help="MongoDB connect URI",
            show_default=True,
        ),
        click.option(
            "-d",
            "--directory",
            default=MongoengineMigrate.default_directory,
            envvar="MONGOENGINE_MIGRATE_DIRECTORY",
            metavar="DIR",
            help="Directory with migrations",
            show_default=True,
        ),
        click.option(
            "-c",
            "--collection",
            default=MongoengineMigrate.default_collection_name,
            envvar="MONGOENGINE_MIGRATE_COLLECTION",
            metavar="COLLECTION",
            help="Collection where schema and state will be stored",
            show_default=True
        ),
        click.option(
            '--mongo-version',
            help="Manually set MongoDB server version. By default it's determined automatically, "
                 "but this requires a permission for 'buildinfo' admin command",
            envvar="MONGOENGINE_MIGRATE_MONGO_VERSION",
            metavar="MONGO_VERSION"
        ),
        click.option(
            '--log-level',
            type=click.Choice(['DEBUG', 'INFO', 'WARNING', 'ERROR'], case_sensitive=False),
            default='INFO',
            envvar="MONGOENGINE_MIGRATE_LOG_LEVEL",
            metavar='LOG_LEVEL',
            help="Logging verbosity level",
            show_default=True
        )
    ]
    for decorator in reversed(decorators):
        f = decorator(f)
    return f


def migration_options(f):
    decorators = [
        click.option(
            '--dry-run',
            default=False,
            is_flag=True,
            help='Dry run mode. Just show queries to be executed, without running migrations'
        ),
        click.option(
            '--schema-only',
            default=False,
            is_flag=True,
            help='Perform migrations without doing any database modifications'
        )
    ]
    for decorator in reversed(decorators):
        f = decorator(f)
    return f


@click.group()
@cli_options
def cli(uri, directory, collection, **kwargs):
    global mongoengine_migrate
    setup_logger(kwargs['log_level'])
    flags.mongo_version = kwargs.get('mongo_version')

    mongoengine_migrate = MongoengineMigrate(mongo_uri=uri,
                                             collection_name=collection,
                                             migrations_dir=directory)
    flags.database2 = mongoengine_migrate.db2


@click.command(short_help='Upgrade db to the given migration')
@click.argument('migration', required=True)
@migration_options
def upgrade(migration, dry_run, schema_only):
    flags.dry_run = dry_run
    flags.schema_only = schema_only

    mongoengine_migrate.upgrade(migration)


@click.command(short_help='Downgrade db to the given migration')
@click.argument('migration', required=True)
@migration_options
def downgrade(migration, dry_run, schema_only):
    flags.dry_run = dry_run
    flags.schema_only = schema_only
    mongoengine_migrate.downgrade(migration)


@click.command(short_help='Migrate db to the given migration. By default is to the last one')
@click.argument('migration', required=False)
@migration_options
def migrate(migration, dry_run, schema_only):
    flags.dry_run = dry_run
    flags.schema_only = schema_only
    mongoengine_migrate.migrate(migration)


@click.command(short_help='Generate migration file based on mongoengine model changes')
@click.option(
    "-m",
    "--models-module",
    default=MongoengineMigrate.default_models_module,
    envvar="MONGOENGINE_MIGRATE_MODELS_MODULE",
    metavar="MODULE_NAME",
    help="Python module where mongoengine models are loaded from",
    show_default=True,
)
def makemigrations(models_module):
    import_module(models_module)
    mongoengine_migrate.makemigrations()


cli.add_command(upgrade)
cli.add_command(downgrade)
cli.add_command(makemigrations)
cli.add_command(migrate)


if __name__ == '__main__':
    try:
        cli()
    except MongoengineMigrateError as e:
        error_handler(e)



import pytest
import itertools
from unittest.mock import Mock
from mongoengine_migrate.action import BaseAction
from mongoengine_migrate.command import BaseCommand
from mongoengine_migrate.exceptions import ActionError


class BaseActionStub(BaseAction):
    COMMANDS_CHAIN = []
    STATE = {}

    def as_schema_patch(self):
        return self.STATE

    def _build_commands_chain(self, old_schema, new_schema):
        return self.COMMANDS_CHAIN


class TestBaseAction:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.new_state = {'new': 'schema'}
        command_attrs = {'forward.return_value': True, 'backward.return_value': True}
        BaseActionStub.COMMANDS_CHAIN = [
            Mock(spec=BaseCommand, **command_attrs),
            Mock(spec=BaseCommand, **command_attrs),
            Mock(spec=BaseCommand, **command_attrs),
            Mock(spec=BaseCommand, **command_attrs)
        ]
        BaseActionStub.STATE = self.new_state
        self.db = Mock()
        self.collection = Mock()
        self.old_state = {'old': 'schema'}

        self.obj = BaseActionStub('base_collection', 'test', 123, test_param=789)

    def test_init__should_set_constructor_params(self):
        assert self.obj.collection_name == 'base_collection'
        assert list(self.obj._init_args) == ['test', 123]
        assert self.obj._init_kwargs == {'test_param': 789}

    def test_prepare__should_store_old_state(self):
        self.obj.prepare(self.old_state)

        assert self.obj.current_schema == self.old_state

    def test_run_forward__on_all_command_successful__should_return_empty_list(self):
        res = self.obj.run_forward(self.db, self.collection)

        assert res == []

    def test_run_forward__with_default_params__should_call_every_command(self):
        self.obj.prepare(self.old_state)
        res = self.obj.run_forward(self.db, self.collection)

        [cmd_mock.forward.assert_called_once_with(self.db,
                                                  self.collection,
                                                  self.old_state,
                                                  self.new_state)
         for cmd_mock in self.obj.COMMANDS_CHAIN]

    @pytest.mark.parametrize('side_effect', (False, None, ActionError))
    def test_run_forward__on_command_fail__should_raise_action_error_and_return_cmd_index(
            self, side_effect
    ):
        failed_index = 1
        self.obj.COMMANDS_CHAIN[failed_index].side_effect = side_effect

        self.obj.prepare(self.old_state)

        with pytest.raises(ActionError) as e:
            self.obj.run_forward(self.db, self.collection)

        assert e.value.args[1] == failed_index

    @pytest.mark.parametrize('side_effect', (False, None, ActionError))
    def test_run_forward__on_command_fail__should_stop_on_failed_command(self, side_effect):
        failed_index = 2
        self.obj.COMMANDS_CHAIN[failed_index].side_effect = side_effect

        self.obj.prepare(self.old_state)

        with pytest.raises(ActionError) as e:
            self.obj.run_forward(self.db, self.collection)

        [cmd_mock.forward.assert_called_once_with(self.db,
                                                  self.collection,
                                                  self.old_state,
                                                  self.new_state)
         for cmd_mock in self.obj.COMMANDS_CHAIN[:failed_index + 1]]
        [cmd_mock.forward.assert_not_called()
         for cmd_mock in self.obj.COMMANDS_CHAIN[failed_index + 1:]]

    @pytest.mark.parametrize('side_effect', (False, None, ActionError))
    def test_run_forward__on_commands_fail_and_ignore_failed__should_return_exceptions(self,
                                                                                       side_effect):
        failed_indexes = (1, 2)
        test_exception = RuntimeError('Something')
        self.obj.COMMANDS_CHAIN[failed_indexes[0]].side_effect = side_effect
        self.obj.COMMANDS_CHAIN[failed_indexes[1]].side_effect = test_exception
        expect = [
            (1, side_effect),
            (2, test_exception)
        ]

        self.obj.prepare(self.old_state)
        res = self.obj.run_forward(self.db, self.collection, ignore_failed=True)

        assert res == expect

    @pytest.mark.parametrize('side_effect', (False, None, ActionError))
    def test_run_forward__on_commands_fail_and_ignore_failed__should_call_all_commands_whatever(
            self,
            side_effect
    ):
        failed_indexes = (1, 2)
        test_exception = RuntimeError('Something')
        self.obj.COMMANDS_CHAIN[failed_indexes[0]].side_effect = side_effect
        self.obj.COMMANDS_CHAIN[failed_indexes[1]].side_effect = test_exception

        self.obj.prepare(self.old_state)
        res = self.obj.run_forward(self.db, self.collection, ignore_failed=True)

        [cmd_mock.forward.assert_called_once_with(self.db,
                                                  self.collection,
                                                  self.old_state,
                                                  self.new_state)
         for cmd_mock in self.obj.COMMANDS_CHAIN]

    def test_run_forward__on_start_from_is_passed__should_run_command_chain_from_this_index(self):
        start_from = 2
        new_state = {'test': 'schema'}

        self.obj.prepare(self.old_state)
        res = self.obj.run_forward(self.db, self.collection, start_from=start_from)

        assert res == []
        [cmd_mock.forward.assert_not_called()
         for cmd_mock in self.obj.COMMANDS_CHAIN[:start_from]]
        [cmd_mock.forward.assert_called_once_with(self.db,
                                                  self.collection,
                                                  self.old_state,
                                                  new_state)
         for cmd_mock in self.obj.COMMANDS_CHAIN[start_from:]]

    @pytest.mark.parametrize('start_from', (-1, 4, 100))
    def test_run_forward__on_wrong_start_from_param__should_raise_value_error(self, start_from):
        self.obj.prepare(self.old_state)
        with pytest.raises(ValueError):
            self.obj.run_forward(self.db, self.collection, start_from=start_from)

    def test_run_backward__on_all_command_successful__should_return_empty_list(self):
        res = self.obj.run_backward(self.db, self.collection)

        assert res == []

    def test_run_backward__with_default_params__should_call_every_command(self):
        self.obj.prepare(self.old_state)
        res = self.obj.run_backward(self.db, self.collection)

        [cmd_mock.backward.assert_called_once_with(self.db,
                                                   self.collection,
                                                   self.old_state,
                                                   self.new_state)
         for cmd_mock in self.obj.COMMANDS_CHAIN]

    @pytest.mark.parametrize('side_effect', (False, None, ActionError))
    def test_run_backward__on_command_fail__should_raise_action_error_and_return_cmd_index(
            self, side_effect
    ):
        failed_index = 1
        self.obj.COMMANDS_CHAIN[failed_index].side_effect = side_effect

        self.obj.prepare(self.old_state)

        with pytest.raises(ActionError) as e:
            self.obj.run_backward(self.db, self.collection)

        assert e.value.args[1] == failed_index

    @pytest.mark.parametrize('side_effect', (False, None, ActionError))
    def test_run_backward__on_command_fail__should_stop_on_failed_command(self, side_effect):
        failed_index = 1
        self.obj.COMMANDS_CHAIN[failed_index].side_effect = side_effect

        self.obj.prepare(self.old_state)

        with pytest.raises(ActionError) as e:
            self.obj.run_backward(self.db, self.collection)

        [cmd_mock.backward.assert_called_once_with(self.db,
                                                   self.collection,
                                                   self.old_state,
                                                   self.new_state)
         for cmd_mock in self.obj.COMMANDS_CHAIN[failed_index:]]
        [cmd_mock.backward.assert_not_called()
         for cmd_mock in self.obj.COMMANDS_CHAIN[:failed_index]]

    @pytest.mark.parametrize('side_effect', (False, None, ActionError))
    def test_run_backward__on_commands_fail_and_ignore_failed__should_return_exceptions(self,
                                                                                       side_effect):
        failed_indexes = (1, 2)
        test_exception = RuntimeError('Something')
        self.obj.COMMANDS_CHAIN[failed_indexes[0]].side_effect = side_effect
        self.obj.COMMANDS_CHAIN[failed_indexes[1]].side_effect = test_exception
        expect = [
            (2, test_exception),
            (1, side_effect)
        ]

        self.obj.prepare(self.old_state)
        res = self.obj.run_backward(self.db, self.collection, ignore_failed=True)

        assert res == expect

    @pytest.mark.parametrize('side_effect', (False, None, ActionError))
    def test_run_backward__on_commands_fail_and_ignore_failed__should_call_all_commands_whatever(
            self,
            side_effect
    ):
        failed_indexes = (1, 2)
        test_exception = RuntimeError('Something')
        self.obj.COMMANDS_CHAIN[failed_indexes[0]].side_effect = side_effect
        self.obj.COMMANDS_CHAIN[failed_indexes[1]].side_effect = test_exception

        self.obj.prepare(self.old_state)
        res = self.obj.run_backward(self.db, self.collection, ignore_failed=True)

        [cmd_mock.backward.assert_called_once_with(self.db,
                                                   self.collection,
                                                   self.old_state,
                                                   self.new_state)
         for cmd_mock in self.obj.COMMANDS_CHAIN]

    def test_run_backward__on_start_from_is_passed__should_run_command_chain_from_this_index(self):
        start_from = 2
        new_state = {'test': 'schema'}

        self.obj.prepare(self.old_state)
        res = self.obj.run_backward(self.db, self.collection, start_from=start_from)

        assert res == []
        [cmd_mock.backward.assert_not_called()
         for cmd_mock in self.obj.COMMANDS_CHAIN[start_from + 1:]]
        [cmd_mock.backward.assert_called_once_with(self.db,
                                                  self.collection,
                                                  self.old_state,
                                                  new_state)
         for cmd_mock in self.obj.COMMANDS_CHAIN[:start_from + 1]]

    @pytest.mark.parametrize('start_from', (-1, 4, 100))
    def test_run_backward__on_wrong_start_from_param__should_raise_value_error(self, start_from):
        self.obj.prepare(self.old_state)
        with pytest.raises(ValueError):
            self.obj.run_backward(self.db, self.collection, start_from=start_from)



__all__ = ['Schema']

from mongoengine_migrate.exceptions import SchemaError


class SchemaAccessMixin:
    """Replace possible KeyError exceptions to SchemaError in dict key
    access methods
    """
    def __access(self, method, item):
        try:
            return method(item)
        except KeyError as e:
            raise SchemaError(f'Unknown key {item!r}') from e

    def __getitem__(self, item):
        return self.__access(super().__getitem__, item)

    def __delitem__(self, key):
        return self.__access(super().__delitem__, key)

    def pop(self, key):
        return self.__access(super().pop, key)

    def popitem(self):
        try:
            return super().popitem()
        except KeyError as e:
            raise SchemaError(f'Schema is empty') from e


class Schema(SchemaAccessMixin, dict):
    """Database schema wrapper"""
    class Document(SchemaAccessMixin, dict):
        class Parameters(SchemaAccessMixin, dict):
            pass

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.__parameters = kwargs.pop('parameters', Schema.Document.Parameters())

        @property
        def parameters(self) -> dict:
            return self.__parameters

        def load(self, document_schema: dict):
            self.__parameters = Schema.Document.Parameters(document_schema.get('parameters', {}))
            self.update(document_schema.get('fields', {}))
            return self

        def dump(self) -> dict:
            return {'fields': dict(self.items()), 'parameters': self.__parameters}

        def __eq__(self, other):
            if self is other:
                return True
            if not isinstance(other, Schema.Document):
                return False

            return self.items() == other.items() and self.parameters == other.parameters

        def __ne__(self, other):
            return not self.__eq__(other)

        def __str__(self):
            return f'Document({super().__repr__()}, parameters={self.parameters!s})'

        def __repr__(self):
            return f'Document({super().__repr__()}, parameters={self.parameters!r})'

    def load(self, db_schema: dict):
        """Load schema from db dict schema representation"""
        self.update({name: Schema.Document().load(schema) for name, schema in db_schema.items()})
        return self

    def dump(self) -> dict:
        """Return schema representation for write to db"""
        return {name: doc.dump() for name, doc in self.items()}

    def __str__(self):
        return f'Schema({super().__repr__()})'

    def __repr__(self):
        return f'Schema({super().__repr__()})'



"""This module contains flags setting on starting, via command line
for example
"""
from typing import Optional
from pymongo.database import Database

#: Dry run mode. Don\'t modify the database and print modification
#: commands which would get executed
dry_run: bool = False


#: Migrate only schema, do not perform any modifications on database
schema_only: bool = False


#: MongoDB server version
mongo_version: Optional[str] = None


#: Another Database object that used for operations which must be
#: performed in a separate connection such as parallel bulk writes
database2: Optional[Database] = None


#: If this prefix contains in collection name then this document
#: is considered as embedded
EMBEDDED_DOCUMENT_NAME_PREFIX = '~'


#: Separator between parent and clild classes in document name string
DOCUMENT_NAME_SEPARATOR = '->'


#: Maximum memory items buffer size on bulk write operations
#: Pay attention: max BSON size is 16Mb
#: https://docs.mongodb.com/manual/reference/limits/#bson-documents
BULK_BUFFER_LENGTH = 10000



__all__ = [
    'CreateField',
    'DropField',
    'AlterField',
    'RenameField'
]

import logging
from typing import Mapping, Any

from pymongo.database import Database

from mongoengine_migrate.exceptions import SchemaError
from ..updater import ByPathContext, ByDocContext, DocumentUpdater
from mongoengine_migrate.schema import Schema
from mongoengine_migrate.utils import document_type_to_class_name
from mongoengine_migrate.graph import MigrationPolicy
from .base import BaseFieldAction

log = logging.getLogger('mongoengine-migrate')


class CreateField(BaseFieldAction):
    """Create field in a given collection"""
    @classmethod
    def build_object(cls,
                     document_type: str,
                     field_name: str,
                     left_schema: Schema,
                     right_schema: Schema):
        match = document_type in left_schema \
                and document_type in right_schema \
                and field_name not in left_schema[document_type] \
                and field_name in right_schema[document_type]
        if match:
            field_params = right_schema[document_type][field_name]
            return cls(document_type=document_type,
                       field_name=field_name,
                       **field_params)

    def to_schema_patch(self, left_schema: Schema):
        keys_to_check = {'type_key', 'db_field'}
        missed_keys = keys_to_check - self.parameters.keys()
        if missed_keys:
            raise SchemaError(f"Missed required parameters in "
                              f"CreateField({self.document_type}...): {missed_keys}")

        # Get schema skeleton for field type
        field_handler_cls = self.get_field_handler_cls(self.parameters['type_key'])
        right_field_schema_skel = field_handler_cls.schema_skel()
        extra_keys = self.parameters.keys() - right_field_schema_skel.keys()
        if extra_keys:
            raise SchemaError(f'Unknown CreateField({self.document_type}...) parameters: '
                              f'{extra_keys}')

        field_params = {
            **right_field_schema_skel,
            **self.parameters
        }
        return [(
            'add',
            self.document_type,
            [(self.field_name, field_params)]
        )]

    def run_forward(self):
        """
        If field is defined as required then force create it with
        default value. Otherwise do nothing since mongoengine creates
        fields automatically on value set
        FIXME: parameters (indexes, acl, etc.)
        """
        def by_path(ctx: ByPathContext):
            # Update documents only
            ctx.collection.update_many(
                {ctx.filter_dotpath: {'$exists': False}, **ctx.extra_filter},
                {'$set': {ctx.update_dotpath: default}}
            )

        def by_doc(ctx: ByDocContext):
            # Update embedded documents
            if db_field not in ctx.document:
                ctx.document[db_field] = default

        db_field = self.parameters['db_field']
        default = self.parameters.get('default')
        is_required = self.parameters.get('required') or self.parameters.get('primary_key')
        if is_required:
            inherit = self._run_ctx['left_schema'][self.document_type].parameters.get('inherit')
            document_cls = document_type_to_class_name(self.document_type) if inherit else None
            updater = DocumentUpdater(self._run_ctx['db'], self.document_type,
                                      self._run_ctx['left_schema'], db_field,
                                      self._run_ctx['migration_policy'], document_cls)
            updater.with_missed_fields().update_combined(by_path, by_doc)

    def run_backward(self):
        """Drop field"""
        def by_path(ctx: ByPathContext):
            ctx.collection.update_many(
                {ctx.filter_dotpath: {'$exists': True}, **ctx.extra_filter},
                {'$unset': {ctx.update_dotpath: ''}},
                array_filters=ctx.build_array_filters()
            )

        db_field = self.parameters['db_field']
        inherit = self._run_ctx['left_schema'][self.document_type].parameters.get('inherit')
        document_cls = document_type_to_class_name(self.document_type) if inherit else None
        updater = DocumentUpdater(self._run_ctx['db'], self.document_type,
                                  self._run_ctx['left_schema'], db_field,
                                  self._run_ctx['migration_policy'], document_cls)
        updater.update_by_path(by_path)

    def prepare(self, db: Database, left_schema: Schema, migration_policy: MigrationPolicy):
        self._prepare(db, left_schema, migration_policy, False)


class DropField(BaseFieldAction):
    """Drop field in a given collection"""
    @classmethod
    def build_object(cls,
                     document_type: str,
                     field_name: str,
                     left_schema: Schema,
                     right_schema: Schema):
        match = document_type in left_schema \
                and document_type in right_schema \
                and field_name in left_schema[document_type] \
                and field_name not in right_schema[document_type]
        if match:
            return cls(document_type=document_type, field_name=field_name)

    def to_schema_patch(self, left_schema: Schema):
        if self.document_type not in left_schema:
            raise SchemaError(f'Document {self.document_type!r} is not in schema')
        if self.field_name not in left_schema[self.document_type]:
            raise SchemaError(f'Field {self.document_type}.{self.field_name} is not in schema')

        left_field_schema = left_schema[self.document_type][self.field_name]

        return [(
            'remove',
            self.document_type,
            [(self.field_name, left_field_schema)]
        )]

    def run_forward(self):
        """Drop field"""
        def by_path(ctx: ByPathContext):
            ctx.collection.update_many(
                {ctx.filter_dotpath: {'$exists': True}, **ctx.extra_filter},
                {'$unset': {ctx.update_dotpath: ''}},
                array_filters=ctx.build_array_filters()
            )

        db_field = self._run_ctx['left_field_schema']['db_field']
        inherit = self._run_ctx['left_schema'][self.document_type].parameters.get('inherit')
        document_cls = document_type_to_class_name(self.document_type) if inherit else None
        updater = DocumentUpdater(self._run_ctx['db'], self.document_type,
                                  self._run_ctx['left_schema'], db_field,
                                  self._run_ctx['migration_policy'], document_cls)
        updater.update_by_path(by_path)

    def run_backward(self):
        """
        If field is defined as required then force create it with
        default value. Otherwise do nothing since mongoengine creates
        fields automatically on value set
        """
        def by_path(ctx: ByPathContext):
            # Update documents only
            ctx.collection.update_many(
                {ctx.filter_dotpath: {'$exists': False}, **ctx.extra_filter},
                {'$set': {ctx.update_dotpath: default}}
            )

        def by_doc(ctx: ByDocContext):
            # Update embedded documents
            if db_field not in ctx.document:
                ctx.document[db_field] = default

        db_field = self._run_ctx['left_field_schema']['db_field']
        default = self._run_ctx['left_field_schema'].get('default')
        is_required = self._run_ctx['left_field_schema'].get('required')
        if is_required:
            inherit = self._run_ctx['left_schema'][self.document_type].parameters.get('inherit')
            document_cls = document_type_to_class_name(self.document_type) if inherit else None
            updater = DocumentUpdater(self._run_ctx['db'], self.document_type,
                                      self._run_ctx['left_schema'], db_field,
                                      self._run_ctx['migration_policy'], document_cls)
            updater.with_missed_fields().update_combined(by_path, by_doc)


class AlterField(BaseFieldAction):
    """Change field parameters or its type, i.e. altering"""
    @classmethod
    def build_object(cls,
                     document_type: str,
                     field_name: str,
                     left_schema: Schema,
                     right_schema: Schema):
        # Check if field still here but its schema has changed
        match = document_type in left_schema \
                and document_type in right_schema \
                and field_name in left_schema[document_type] \
                and field_name in right_schema[document_type] \
                and left_schema[document_type][field_name] != right_schema[document_type][field_name]
        if match:
            # Consider items which was changed and added, skip those
            # ones which was unchanged or was removed
            right_field_schema = right_schema[document_type][field_name]
            left_field_schema = left_schema[document_type][field_name]
            action_params = dict(right_field_schema.items() - left_field_schema.items())
            # FIXME: use function below
            # field_params = cls._fix_field_params(document_type,
            #                                      field_name,
            #                                      field_params,
            #                                      old_schema,
            #                                      new_schema)
            return cls(document_type=document_type, field_name=field_name, **action_params)

    def to_schema_patch(self, left_schema: Schema):
        if self.document_type not in left_schema:
            raise SchemaError(f'Document {self.document_type!r} is not in schema')
        if self.field_name not in left_schema[self.document_type]:
            raise SchemaError(f'Field {self.document_type}.{self.field_name} is not in schema')

        left_field_schema = left_schema[self.document_type][self.field_name]

        # Get schema skeleton for field type
        field_handler_cls = self.get_field_handler_cls(
            self.parameters.get('type_key', left_field_schema['type_key'])
        )
        right_schema_skel = field_handler_cls.schema_skel()
        extra_keys = self.parameters.keys() - right_schema_skel.keys()
        if extra_keys:
            raise SchemaError(f'Unknown keys in schema of field '
                              f'{self.document_type}.{self.field_name}: {extra_keys}')

        # Shortcuts
        left = left_field_schema
        params = self.parameters

        # Remove params
        d = [('remove', f'{self.document_type}.{self.field_name}', [(key, ())])
             for key in left.keys() - right_schema_skel.keys()]
        # Add new params
        d += [('add', f'{self.document_type}.{self.field_name}', [(key, params[key])])
              for key in params.keys() - left.keys()]
        # Change params if they are requested to be changed
        d += [('change',
               f'{self.document_type}.{self.field_name}.{key}',
               (left[key], params[key]))
              for key in params.keys() & left.keys()
              if left[key] != params[key]]

        return d

    def run_forward(self):
        self._run_migration(self._run_ctx['left_field_schema'], self.parameters, swap=False)

    def run_backward(self):
        self._run_migration(self._run_ctx['left_field_schema'], self.parameters, swap=True)

    def _run_migration(self, left_field_schema: dict, parameters: Mapping[str, Any], swap=False):
        type_key = left_field_schema['type_key']
        field_handler_cls = self.get_field_handler_cls(type_key)
        skel = field_handler_cls.schema_skel()
        right_field_schema = {k: parameters.get(k, left_field_schema.get(k, v))
                              for k, v in skel.items()}

        if swap:
            left_field_schema, right_field_schema = right_field_schema, left_field_schema
            type_key = left_field_schema['type_key']

        field_handler = self._get_field_handler(type_key, left_field_schema, right_field_schema)
        db_field = left_field_schema['db_field']

        # Change field type first, obtain new field handler object
        # and process other parameters with it
        if type_key != right_field_schema['type_key']:
            log.debug(">> Change 'type_key': %s => %s",
                      repr(type_key),
                      repr(right_field_schema['type_key']))
            field_handler.change_param(db_field, 'type_key')
            field_handler = self._get_field_handler(right_field_schema['type_key'],
                                                    left_field_schema,
                                                    right_field_schema)

        # Try to process all parameters on same order to avoid
        # potential problems on repeated launches if some query on
        # previous launch was failed
        unset = object()
        for name, new_value in sorted(right_field_schema.items()):
            old_value = left_field_schema.get(name, unset)
            if name == 'type_key' or new_value == old_value:
                continue

            log.debug(">> Change %s: %s => %s", repr(name), repr(old_value), repr(new_value))
            field_handler.change_param(db_field, name)

            # If `db_field` was changed then work with new name further
            if name == 'db_field':
                db_field = right_field_schema['db_field']

    @classmethod
    def _fix_field_params(cls,
                          document_type: str,
                          field_name: str,
                          field_params: Mapping[str, Any],
                          old_schema: Schema,
                          new_schema: Schema) -> Mapping[str, Any]:
        """
        Search for potential problems which could be happened during
        migration and return fixed field schema. If such problem
        could not be resolved only by changing parameters then raise
        an SchemaError
        :param document_type:
        :param field_name:
        :param field_params:
        :param old_schema:
        :param new_schema:
        :raises SchemaError: when some problem found
        :return:
        """
        # TODO: Check all defaults in diffs against choices, required, etc.
        # TODO: check nones for type_key, etc.
        new_changes = {k: v.new for k, v in field_params.items()}

        # Field becomes required or left as required without changes
        become_required = new_changes.get('required',
                                          new_schema.get(field_name, {}).get('required'))

        # 'default' diff object has default parameter or field become
        # field with default or field has default value already
        default = (field_params.get('required') and field_params['required'].default) \
            or new_changes.get('default') \
            or new_schema.get(field_name, {}).get('default')
        if become_required and default is None:
            # TODO: replace following error on interactive mode
            raise SchemaError(f'Field {document_type}.{field_name} could not be '
                              f'created since it defined as required but has not a default value')

        return field_params


class RenameField(BaseFieldAction):
    """Rename field"""
    priority = 10

    #: How much percent of items in schema diff of two fields in the
    #: same collection should be equal to consider such change as
    #: field rename instead of drop/create
    similarity_threshold = 70

    def __init__(self, document_type: str, field_name: str, *, new_name, **kwargs):
        super().__init__(document_type, field_name, new_name=new_name, **kwargs)
        self.new_name = new_name

    @classmethod
    def build_object(cls,
                     document_type: str,
                     field_name: str,
                     left_schema: Schema,
                     right_schema: Schema):
        # Check if field exists under different name in schema
        # Field also can have small schema changes in the same time
        # So we try to get similarity percentage and if it more than
        # threshold then we're consider such change as rename/alter.
        # Otherwise it is drop/create
        match = document_type in left_schema \
                and document_type in right_schema \
                and field_name in left_schema[document_type] \
                and field_name not in right_schema[document_type]
        if not match:
            return

        left_field_schema = left_schema[document_type][field_name]
        candidates = []
        for right_field_name, right_field_schema in right_schema[document_type].items():
            # Skip fields which was not renamed
            # Changing 'db_field' parameter is altering, not renaming
            if right_field_name in left_schema[document_type]:
                continue

            # Model field renamed, but db field is the same
            db_field = right_field_schema.get('db_field')
            if db_field == right_field_name:
                candidates = [(right_field_name, right_field_schema)]
                break

            # Take only common keys to estimate similarity
            # 'type_key' may get changed which means that change of one
            # key leads to many changes in schema. These changes
            # should not be considered as valueable
            keys = left_field_schema.keys() & right_field_schema.keys() - {'db_field'}
            if keys:
                p = sum(left_field_schema[k] == right_field_schema[k] for k in keys) / len(keys)
                if p * 100 >= cls.similarity_threshold:
                    candidates.append((right_field_name, right_field_schema))

        if len(candidates) == 1:
            return cls(document_type=document_type,
                       field_name=field_name,
                       new_name=candidates[0][0])

    def to_schema_patch(self, left_schema: Schema):
        if self.document_type not in left_schema:
            raise SchemaError(f'Document {self.document_type!r} is not in schema')
        if self.field_name not in left_schema[self.document_type]:
            raise SchemaError(f'Field {self.document_type}.{self.field_name} is not in schema')

        left_field_schema = left_schema[self.document_type][self.field_name]

        return [
            ('remove', f'{self.document_type}', [(self.field_name, left_field_schema)]),
            ('add', f'{self.document_type}', [(self.new_name, left_field_schema)])
        ]

    def run_forward(self):
        """Renaming mongoengine model field does not required some
        db changes. It's supposed that possible `db_field` modification,
        which could exist during field renaming, is handled by
        AlterField action
        """

    def run_backward(self):
        """Renaming mongoengine model field does not required some
        db changes. It's supposed that possible `db_field` modification,
        which could exist during field renaming, is handled by
        AlterField action
        """



__all__ = [
    'MongoengineMigrateError',
    'MigrationGraphError',
    'ActionError',
    'SchemaError',
    'MigrationError',
    'InconsistencyError'
]


class MongoengineMigrateError(Exception):
    """Generic error"""


class MigrationGraphError(MongoengineMigrateError):
    """Error related to migration modules names, dependencies, etc."""


class ActionError(MongoengineMigrateError):
    """Error related to Action itself"""


class SchemaError(MongoengineMigrateError):
    """Error related to schema errors"""


class MigrationError(MongoengineMigrateError):
    """Error which could occur during migration"""


class InconsistencyError(MigrationError):
    """Error which could occur during migration if data inconsistency
    was detected
    """



__all__ = [
    'HistoryCallKind',
    'CollectionQueryTracer',
    'DatabaseQueryTracer'
]

import logging
from enum import Enum
from typing import NamedTuple, Dict, Tuple, Any

import wrapt
from bson import ObjectId
from pymongo.collection import Collection

_sentinel = object()


log = logging.getLogger('mongoengine-migrate')


class HistoryCallKind(Enum):
    READ = 'READ'
    MODIFY = "MODIFY"
    AGGREGATE = "AGGREGATE"


class InsertOneResultMock(NamedTuple):
    inserted_id: ObjectId = ObjectId('000000000000000000000000')


class InsertManyResultMock(NamedTuple):
    inserted_ids: Tuple[ObjectId] = (ObjectId('000000000000000000000000'), )


class UpdateResultMock(NamedTuple):
    raw_result: Dict[str, Any] = {}
    matched_count: int = 0
    modified_count: int = 0
    upserted_id: ObjectId = ObjectId('000000000000000000000000')


class DeleteResultMock(NamedTuple):
    raw_result: Dict[str, Any] = {}
    deleted_count: int = 0


class BulkWriteResultMock(NamedTuple):
    bulk_api_result: Dict[str, Any] = {}
    inserted_count: int = 0
    matched_count: int = 0
    modified_count: int = 0
    deleted_count: int = 0
    upserted_count: int = 0
    upserted_ids: Tuple[ObjectId] = (ObjectId('000000000000000000000000'), )


def make_history_method(func_name, method_kind, return_value=_sentinel):
    def w(instance, *args, **kwargs):
        args_str = ', '.join(f'\n  {arg}' for arg in args)
        kwargs_str = ', '.join(f"\n  {name}={val}" for name, val in sorted(kwargs.items()))
        arguments = f'{args_str}{"," if kwargs_str else ""}{kwargs_str}'
        if arguments:
            arguments += '\n'
        collection_name = instance.__wrapped__.full_name
        log.info('* %s.%s(%s)', collection_name, func_name, arguments)

        if return_value == _sentinel:
            f = getattr(instance.__wrapped__, func_name)
            return f(*args, **kwargs)

        return return_value
    return w


class CollectionQueryTracer(wrapt.ObjectProxy):
    """
    pymongo.Collection wrapper object which mocks modification methods
    calls and writes their call to history
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # Collection modification methods
    bulk_write = make_history_method('bulk_write', 'MODIFY', return_value=BulkWriteResultMock())
    insert_one = make_history_method('insert_one', 'MODIFY', return_value=InsertOneResultMock())
    insert_many = make_history_method('insert_many', 'MODIFY', return_value=InsertManyResultMock())
    replace_one = make_history_method('replace_one', 'MODIFY', return_value=UpdateResultMock())
    update_one = make_history_method('update_one', 'MODIFY', return_value=UpdateResultMock())
    update_many = make_history_method('update_many', 'MODIFY', return_value=UpdateResultMock())
    drop = make_history_method('drop', 'MODIFY', return_value=None)
    delete_one = make_history_method('delete_one', 'MODIFY', return_value=DeleteResultMock())
    delete_many = make_history_method('delete_many', 'MODIFY', return_value=DeleteResultMock())
    create_indexes = make_history_method('create_indexes', 'MODIFY', return_value=[])
    create_index = make_history_method('create_index', 'MODIFY', return_value=None)
    ensure_index = make_history_method('ensure_index', 'MODIFY', return_value=None)
    drop_indexes = make_history_method('drop_indexes', 'MODIFY', return_value=None)
    drop_index = make_history_method('drop_index', 'MODIFY', return_value=None)
    reindex = make_history_method('reindex', 'MODIFY', return_value=None)
    rename = make_history_method('rename', 'MODIFY', return_value=None)
    find_one_and_delete = make_history_method('find_one_and_delete', 'MODIFY', return_value=None)
    find_one_and_replace = make_history_method('find_one_and_replace', 'MODIFY', return_value=None)
    find_one_and_update = make_history_method('find_one_and_update', 'MODIFY', return_value=None)
    save = make_history_method('save', 'MODIFY', return_value=ObjectId('000000000000000000000000'))
    insert = make_history_method('insert', 'MODIFY', return_value=[])
    update = make_history_method('update', 'MODIFY', return_value=[])
    remove = make_history_method('remove', 'MODIFY', return_value=None)
    find_and_modify = make_history_method('find_and_modify', 'MODIFY', return_value=None)

    # Aggregation methods
    aggregate = make_history_method('aggregate', 'AGGREGATE', return_value=[])
    aggregate_raw_batches = make_history_method('aggregate_raw_batches',
                                                'AGGREGATE',
                                                return_value=[])

    # Collection reading methods
    find_one = make_history_method('find_one', 'READ')
    find = make_history_method('find', 'READ')
    find_raw_batches = make_history_method('find_raw_batches', 'READ')
    estimated_document_count = make_history_method('estimated_document_count', 'READ')
    count_documents = make_history_method('count_documents', 'READ')
    list_indexes = make_history_method('list_indexes', 'READ')
    index_information = make_history_method('index_information', 'READ')
    options = make_history_method('options', 'READ')
    watch = make_history_method('watch', 'READ')
    group = make_history_method('group', 'READ')
    distinct = make_history_method('distinct', 'READ')
    map_reduce = make_history_method('map_reduce', 'READ')
    inline_map_reduce = make_history_method('inline_map_reduce', 'READ')


class DatabaseQueryTracer(wrapt.ObjectProxy):
    """pymongo.Database wrapper which is acting as original object,
    but returns CollectionQueryTracer object instead of Collection
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __getitem__(self, item):
        col = super().__getitem__(item)
        return CollectionQueryTracer(col)

    def __getattr__(self, item):
        val = super().__getattr__(item)
        if isinstance(val, Collection):
            return CollectionQueryTracer(val)

        return val

    def get_collection(self, *args, **kwargs):
        col = super().get_collection(*args, **kwargs)
        return CollectionQueryTracer(col)

    def create_collection(self, *args, **kwargs):
        col = super().create_collection(*args, **kwargs)
        return CollectionQueryTracer(col)



import jsonpath_rw
from bson import ObjectId
from mongoengine import Document, fields, EmbeddedDocument

from mongoengine_migrate.schema import Schema


def get_classes():
    class Schema1EmbDoc1(EmbeddedDocument):
        embdoc1_int = fields.IntField()
        embdoc1_str = fields.StringField()
        embdoc1_str_empty = fields.StringField()
        embdoc1_str_ten = fields.StringField(choices=[str(x) for x in range(11)])
        embdoc1_float = fields.FloatField()
        embdoc1_int_empty = fields.IntField()
        embdoc1_long = fields.LongField()
        embdoc1_decimal = fields.DecimalField()
        embdoc1_complex_datetime = fields.ComplexDateTimeField()
        embdoc1_list = fields.ListField()
        embdoc1_ref_doc1 = fields.ReferenceField('Schema1Doc1')
        embdoc1_emb_embdoc1 = fields.EmbeddedDocumentField('self')
        embdoc1_emblist_embdoc1 = fields.EmbeddedDocumentListField('self')
        embdoc1_emb_embdoc2 = fields.EmbeddedDocumentField('Schema1EmbDoc2')
        embdoc1_emblist_embdoc2 = fields.EmbeddedDocumentListField('Schema1EmbDoc2')

    class Schema1EmbDoc2(EmbeddedDocument):
        embdoc2_int = fields.IntField()
        embdoc2_str = fields.StringField()
        embdoc2_str_empty = fields.StringField()
        embdoc2_str_ten = fields.StringField(choices=[str(x) for x in range(11)])
        embdoc2_float = fields.FloatField()
        embdoc2_int_empty = fields.IntField()
        embdoc2_long = fields.LongField()
        embdoc2_decimal = fields.DecimalField()
        embdoc2_complex_datetime = fields.ComplexDateTimeField()
        embdoc2_list = fields.ListField()
        embdoc2_ref_doc1 = fields.ReferenceField('Schema1Doc1')
        embdoc2_emb_embdoc2 = fields.EmbeddedDocumentField('self')
        embdoc2_emblist_embdoc2 = fields.EmbeddedDocumentListField('self')

    class Schema1Doc1(Document):
        doc1_int = fields.IntField()
        doc1_str = fields.StringField()
        doc1_str_empty = fields.StringField()
        doc1_str_ten = fields.StringField(choices=[str(x) for x in range(11)])
        doc1_float = fields.FloatField()
        doc1_int_empty = fields.IntField()
        doc1_long = fields.LongField()
        doc1_decimal = fields.DecimalField()
        doc1_complex_datetime = fields.ComplexDateTimeField()
        doc1_list = fields.ListField()
        doc1_ref_self = fields.ReferenceField('self')
        doc1_cachedref_self = fields.CachedReferenceField('self')
        doc1_emb_embdoc1 = fields.EmbeddedDocumentField('Schema1EmbDoc1')
        doc1_emb_embdoc2 = fields.EmbeddedDocumentField('Schema1EmbDoc2')
        doc1_emblist_embdoc1 = fields.EmbeddedDocumentListField('Schema1EmbDoc1')
        doc1_emblist_embdoc2 = fields.EmbeddedDocumentListField('Schema1EmbDoc2')

    return Schema1EmbDoc1, Schema1EmbDoc2, Schema1Doc1


def get_schema():
    schema = Schema({
        '~Schema1EmbDoc1': Schema.Document(
            {
                'embdoc1_int': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': None,
                    'sparse': False, 'unique': False, 'required': False, 'db_field': 'embdoc1_int',
                    'primary_key': False, 'type_key': 'IntField', 'max_value': None,
                    'min_value': None},
                'embdoc1_str': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': None,
                    'sparse': False, 'unique': False, 'required': False, 'db_field': 'embdoc1_str',
                    'primary_key': False, 'type_key': 'StringField', 'max_length': None,
                    'regex': None, 'min_length': None},
                'embdoc1_str_empty': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': None,
                    'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'embdoc1_str_empty',
                    'primary_key': False, 'type_key': 'StringField', 'max_length': None,
                    'regex': None, 'min_length': None},
                'embdoc1_str_ten': {
                    'unique_with': None, 'null': False,
                    'choices': ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '10'],
                    'default': None, 'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'embdoc1_str_ten', 'primary_key': False, 'type_key': 'StringField',
                    'max_length': None, 'regex': None, 'min_length': None},
                'embdoc1_float': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': None,
                    'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'embdoc1_float', 'primary_key': False, 'type_key': 'FloatField',
                    'max_value': None, 'min_value': None},
                'embdoc1_long': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': None,
                    'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'embdoc1_long', 'primary_key': False, 'type_key': 'LongField',
                    'max_value': None, 'min_value': None},
                'embdoc1_int_empty': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': None,
                    'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'embdoc1_int_empty', 'primary_key': False, 'type_key': 'IntField',
                    'max_value': None, 'min_value': None},
                'embdoc1_decimal': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': None,
                    'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'embdoc1_decimal', 'primary_key': False,
                    'type_key': 'DecimalField', 'max_value': None, 'min_value': None,
                    'force_string': False, 'precision': 2, 'rounding': 'ROUND_HALF_UP'},
                'embdoc1_complex_datetime': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': None,
                    'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'embdoc1_complex_datetime', 'primary_key': False,
                    'type_key': 'ComplexDateTimeField', 'max_length': None,
                    'regex': None, 'min_length': None, 'separator': '.'},
                'embdoc1_list': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': [],
                    'sparse': False, 'unique': False, 'required': False, 'db_field': 'embdoc1_list',
                    'primary_key': False, 'type_key': 'ListField', 'max_length': None},
                'embdoc1_ref_doc1': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': [],
                    'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'embdoc1_ref_doc1', 'primary_key': False,
                    'type_key': 'ReferenceField', 'target_doctype': 'Schema1Doc1', 'dbref': False},
                'embdoc1_emb_embdoc1': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': None,
                    'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'embdoc1_emb_embdoc1', 'primary_key': False,
                    'type_key': 'EmbeddedDocumentField', 'target_doctype': '~Schema1EmbDoc1'},
                'embdoc1_emblist_embdoc1': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': [],
                    'sparse': False,'unique': False, 'required': False,
                    'db_field': 'embdoc1_emblist_embdoc1', 'primary_key': False,
                    'type_key': 'EmbeddedDocumentListField', 'target_doctype': '~Schema1EmbDoc1',
                    'max_length': None},
                'embdoc1_emb_embdoc2': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': None,
                    'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'embdoc1_emb_embdoc2', 'primary_key': False,
                    'type_key': 'EmbeddedDocumentField', 'target_doctype': '~Schema1EmbDoc2'},
                'embdoc1_emblist_embdoc2': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': [],
                    'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'embdoc1_emblist_embdoc2', 'primary_key': False,
                    'type_key': 'EmbeddedDocumentListField', 'target_doctype': '~Schema1EmbDoc2',
                    'max_length': None}
            },
            parameters={}
        ),
        '~Schema1EmbDoc2': Schema.Document(
            {
                'embdoc2_int': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': None,
                    'sparse': False, 'unique': False, 'required': False, 'db_field': 'embdoc2_int',
                    'primary_key': False, 'type_key': 'IntField', 'max_value': None,
                    'min_value': None},
                'embdoc2_str': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': None,
                    'sparse': False, 'unique': False, 'required': False, 'db_field': 'embdoc2_str',
                    'primary_key': False, 'type_key': 'StringField', 'max_length': None,
                    'regex': None, 'min_length': None},
                'embdoc2_str_empty': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': None,
                    'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'embdoc2_str_empty',
                    'primary_key': False, 'type_key': 'StringField', 'max_length': None,
                    'regex': None, 'min_length': None},
                'embdoc2_str_ten': {
                    'unique_with': None, 'null': False,
                    'choices': ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '10'],
                    'default': None, 'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'embdoc2_str_ten', 'primary_key': False, 'type_key': 'StringField',
                    'max_length': None, 'regex': None, 'min_length': None},
                'embdoc2_float': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': None,
                    'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'embdoc2_float', 'primary_key': False, 'type_key': 'FloatField',
                    'max_value': None, 'min_value': None},
                'embdoc2_long': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': None,
                    'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'embdoc2_long', 'primary_key': False, 'type_key': 'LongField',
                    'max_value': None, 'min_value': None},
                'embdoc2_int_empty': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': None,
                    'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'embdoc2_int_empty', 'primary_key': False, 'type_key': 'IntField',
                    'max_value': None, 'min_value': None},
                'embdoc2_decimal': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': None,
                    'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'embdoc2_decimal', 'primary_key': False,
                    'type_key': 'DecimalField', 'max_value': None, 'min_value': None,
                    'force_string': False, 'precision': 2, 'rounding': 'ROUND_HALF_UP'},
                'embdoc2_complex_datetime': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': None,
                    'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'embdoc2_complex_datetime', 'primary_key': False,
                    'type_key': 'ComplexDateTimeField', 'max_length': None,
                    'regex': None, 'min_length': None, 'separator': '.'},
                'embdoc2_list': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': [],
                    'sparse': False, 'unique': False, 'required': False, 'db_field': 'embdoc2_list',
                    'primary_key': False, 'type_key': 'ListField', 'max_length': None},
                'embdoc2_ref_doc1': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': [],
                    'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'embdoc2_ref_doc1', 'primary_key': False,
                    'type_key': 'ReferenceField', 'target_doctype': 'Schema1Doc1', 'dbref': False},
                'embdoc2_emb_embdoc2': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': None,
                    'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'embdoc2_emb_embdoc2', 'primary_key': False,
                    'type_key': 'EmbeddedDocumentField', 'target_doctype': '~Schema1EmbDoc2'},
                'embdoc2_emblist_embdoc2': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': [],
                    'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'embdoc2_emblist_embdoc2', 'primary_key': False,
                    'type_key': 'EmbeddedDocumentListField', 'target_doctype': '~Schema1EmbDoc2',
                    'max_length': None}
            },
            parameters={}
        ),
        'Schema1Doc1': Schema.Document(
            {
                'doc1_int': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': None,
                    'sparse': False, 'unique': False, 'required': False, 'db_field': 'doc1_int',
                    'primary_key': False, 'type_key': 'IntField', 'max_value': None,
                    'min_value': None},
                'doc1_str': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': None,
                    'sparse': False, 'unique': False, 'required': False, 'db_field': 'doc1_str',
                    'primary_key': False, 'type_key': 'StringField', 'max_length': None,
                    'regex': None, 'min_length': None},
                'doc1_str_empty': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': None,
                    'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'doc1_str_empty',
                    'primary_key': False, 'type_key': 'StringField', 'max_length': None,
                    'regex': None, 'min_length': None},
                'doc1_str_ten': {
                    'unique_with': None, 'null': False,
                    'choices': ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '10'],
                    'default': None, 'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'doc1_str_ten', 'primary_key': False, 'type_key': 'StringField',
                    'max_length': None, 'regex': None, 'min_length': None},
                'doc1_float': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': None,
                    'sparse': False, 'unique': False, 'required': False, 'db_field': 'doc1_float',
                    'primary_key': False, 'type_key': 'FloatField', 'max_value': None,
                    'min_value': None},
                'doc1_long': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': None,
                    'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'doc1_long', 'primary_key': False, 'type_key': 'LongField',
                    'max_value': None, 'min_value': None},
                'doc1_int_empty': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': None,
                    'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'doc1_int_empty', 'primary_key': False, 'type_key': 'IntField',
                    'max_value': None, 'min_value': None},
                'doc1_decimal': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': None,
                    'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'doc1_decimal', 'primary_key': False,
                    'type_key': 'DecimalField', 'max_value': None, 'min_value': None,
                    'force_string': False, 'precision': 2, 'rounding': 'ROUND_HALF_UP'},
                'doc1_complex_datetime': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': None,
                    'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'doc1_complex_datetime', 'primary_key': False,
                    'type_key': 'ComplexDateTimeField', 'max_length': None,
                    'regex': None, 'min_length': None, 'separator': '.'},
                'doc1_list': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': [],
                    'sparse': False, 'unique': False, 'required': False, 'db_field': 'doc1_list',
                    'primary_key': False, 'type_key': 'ListField', 'max_length': None},
                'doc1_ref_self': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': [],
                    'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'doc1_ref_self', 'primary_key': False,
                    'type_key': 'ReferenceField', 'target_doctype': 'self', 'dbref': False},
                'doc1_cachedref_self': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': [],
                    'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'doc1_cachedref_self', 'primary_key': False,
                    'type_key': 'CachedReferenceField', 'target_doctype': 'self', 'fields': []},
                'doc1_emb_embdoc1': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': None,
                    'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'doc1_emb_embdoc1', 'primary_key': False,
                    'type_key': 'EmbeddedDocumentField', 'target_doctype': '~Schema1EmbDoc1'},
                'doc1_emb_embdoc2': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': None,
                    'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'doc1_emb_embdoc2', 'primary_key': False,
                    'type_key': 'EmbeddedDocumentField', 'target_doctype': '~Schema1EmbDoc2'},
                'doc1_emblist_embdoc1': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': [],
                    'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'doc1_emblist_embdoc1', 'primary_key': False,
                    'type_key': 'EmbeddedDocumentListField', 'target_doctype': '~Schema1EmbDoc1',
                    'max_length': None},
                'doc1_emblist_embdoc2': {
                    'unique_with': None, 'null': False, 'choices': None, 'default': [],
                    'sparse': False, 'unique': False, 'required': False,
                    'db_field': 'doc1_emblist_embdoc2', 'primary_key': False,
                    'type_key': 'EmbeddedDocumentListField',  'target_doctype': '~Schema1EmbDoc2',
                    'max_length': None}
            },
            parameters={'collection': 'schema1_doc1'}
        )
    })

    return schema


def setup_db():
    Schema1EmbDoc1, Schema1EmbDoc2, Schema1Doc1 = get_classes()

    # doc1
    n = 1
    doc1_1 = Schema1Doc1(id=ObjectId(f'{n:024}'),
                         doc1_int=n,
                         doc1_str=f'str{n}',
                         doc1_str_ten=str(n % 10),
                         doc1_list=[n * 10000 + 1, f'str{n * 10000 + 1}', None]).save()

    # doc1.embdoc1
    n = 2
    embdoc2_1 = Schema1EmbDoc1(embdoc1_int=n * 10 + 1,
                               embdoc1_str=f'str{n * 10 + 1}',
                               embdoc1_str_ten=str(n % 10),
                               embdoc1_list=[
                                   n * 10000 + 1001, f'str{n * 10000 + 1001}', None
                               ])
    doc1_1 = Schema1Doc1(id=ObjectId(f'{n:024}'),
                         doc1_int=n,
                         doc1_str=f'str{n}',
                         doc1_str_ten=str(n % 10),
                         doc1_emb_embdoc1=embdoc2_1).save()

    # doc1.[embdoc1]
    n = 3
    embdoc1_x = [
        Schema1EmbDoc1(embdoc1_int=n * 10 + 1,
                       embdoc1_str=f'str{n * 10 + 1}',
                       embdoc1_str_ten=str(n % 10),
                       embdoc1_list=[n * 10000 + 1001, f'str{n * 10000 + 1001}', None]),
        Schema1EmbDoc1(embdoc1_int=n * 10 + 2,
                       embdoc1_str=f'str{n * 10 + 2}',
                       embdoc1_str_ten=str(n % 10),
                       embdoc1_list=[n * 10000 + 2001, f'str{n * 10000 + 2001}', None]),
    ]
    doc1_1 = Schema1Doc1(id=ObjectId(f'{n:024}'),
                         doc1_int=n,
                         doc1_str=f'str{n}',
                         doc1_str_ten=str(n % 10),
                         doc1_emblist_embdoc1=embdoc1_x).save()

    # doc1.embdoc1.[embdoc2]
    n = 4
    embdoc2_x = [
        Schema1EmbDoc2(embdoc2_int=n * 100 + 11,
                       embdoc2_str=f'str{n * 100 + 11}',
                       embdoc2_str_ten=str(n % 10),
                       embdoc2_list=[n * 100 + 11, f'str{n * 100 + 11}', None]),
        Schema1EmbDoc2(embdoc2_int=n * 100 + 12,
                       embdoc2_str=f'str{n * 100 + 12}',
                       embdoc2_str_ten=str(n % 10),
                       embdoc2_list=[n * 100 + 11, f'str{n * 100 + 11}', None]),
    ]
    embdoc2_1 = Schema1EmbDoc1(embdoc1_int=n * 10 + 1,
                               embdoc1_str=f'str{n * 10 + 1}',
                               embdoc1_str_ten=str(n % 10),
                               embdoc1_emblist_embdoc2=embdoc2_x)
    doc1_1 = Schema1Doc1(id=ObjectId(f'{n:024}'),
                         doc1_int=n,
                         doc1_str=f'str{n}',
                         doc1_str_ten=str(n % 10),
                         doc1_emb_embdoc1=embdoc2_1).save()

    items = (
        ('embdoc1', Schema1EmbDoc1, 'embdoc2', Schema1EmbDoc2),
        ('embdoc1', Schema1EmbDoc1, 'embdoc1', Schema1EmbDoc1),
    )
    for field1, cls1, field2, cls2 in items:
        # doc1.[embdoc1].[embdoc1/embdoc2]
        n += 1  # 5, 7
        embdoc2_1 = [
            cls2(**{
                f'{field2}_int': n * 100 + 11,
                f'{field2}_str': f'str{n * 100 + 11}',
                f'{field2}_str_ten': str(n % 10),
                f'{field2}_list': [n * 10000 + 1101, f'str{n * 10000 + 1101}', None]
            }),
            cls2(**{
                f'{field2}_int': n * 100 + 12,
                f'{field2}_str': f'str{n * 100 + 12}',
                f'{field2}_str_ten': str(n % 10),
                f'{field2}_list': [n * 10000 + 1201, f'str{n * 10000 + 1201}', None]
            })
        ]
        embdoc2_2 = [
            cls2(**{
                f'{field2}_int': n * 100 + 21,
                f'{field2}_str': f'str{n * 100 + 21}',
                f'{field2}_str_ten': str(n % 10),
                f'{field2}_list': [n * 10000 + 2101, f'str{n * 10000 + 2101}', None]
            }),
            cls2(**{
                f'{field2}_int': n * 100 + 22,
                f'{field2}_str': f'str{n * 100 + 22}',
                f'{field2}_str_ten': str(n % 10),
                f'{field2}_list': [n * 10000 + 2201, f'str{n * 10000 + 2201}', None]
            })
        ]
        embdoc1_x = [
            cls1(**{
                f'{field1}_int': n * 10 + 1,
                f'{field1}_str': f'str{n * 10 + 1}',
                f'{field1}_str_ten': str(n % 10),
                f'{field1}_emblist_{field2}': embdoc2_1
            }),
            cls1(**{
                f'{field1}_int': n * 10 + 2,
                f'{field1}_str': f'str{n * 10 + 2}',
                f'{field1}_str_ten': str(n % 10),
                f'{field1}_emblist_{field2}': embdoc2_2
            })
        ]
        doc1_1 = Schema1Doc1(**{f'id': ObjectId(f'{n:024}'),
                                f'doc1_int': n,
                                f'doc1_str': f'str{n}',
                                f'doc1_str_ten': str(n % 10),
                                f'doc1_emblist_{field1}': embdoc1_x}).save()

        # doc1.[embdoc1].embdoc1.[embdoc1/embdoc2]
        n += 1  # 6, 8
        embdoc2_1 = [
            cls2(**{
                f'{field2}_int': n * 1000 + 111,
                f'{field2}_str': f'str{n * 1000 + 111}',
                f'{field2}_str_ten': str(n % 10),
                f'{field2}_list': [n * 10000 + 1111, f'str{n * 10000 + 1111}', None]
            }),
            cls2(**{
                f'{field2}_int': n * 1000 + 112,
                f'{field2}_str': f'str{n * 1000 + 112}',
                f'{field2}_str_ten': str(n % 10),
                f'{field2}_list': [n * 10000 + 1121, f'str{n * 10000 + 1121}', None]
            })
        ]
        embdoc2_2 = [
            cls2(**{
                f'{field2}_int': n * 1000 + 211,
                f'{field2}_str': f'str{n * 1000 + 211}',
                f'{field2}_str_ten': str(n % 10),
                f'{field2}_list': [n * 10000 + 2111, f'str{n * 10000 + 2111}', None]
            }),
            cls2(**{
                f'{field2}_int': n * 1000 + 212,
                f'{field2}_str': f'str{n * 1000 + 212}',
                f'{field2}_str_ten': str(n % 10),
                f'{field2}_list': [n * 10000 + 2121, f'str{n * 10000 + 2121}', None]
            })
        ]
        embdoc1_1 = cls1(**{
            f'{field1}_int': n * 100 + 11,
            f'{field1}_str': f'str{n * 100 + 11}',
            f'{field1}_str_ten': str(n % 10),
            f'{field1}_emblist_{field2}': embdoc2_1
        })
        embdoc1_2 = cls1(**{
            f'{field1}_int': n * 100 + 21,
            f'{field1}_str': f'str{n * 100 + 21}',
            f'{field1}_str_ten': str(n % 10),
            f'{field1}_emblist_{field2}': embdoc2_2
        })
        embdoc1_x = [
            cls1(**{
                f'{field1}_int': n * 10 + 1,
                f'{field1}_str': f'str{n * 10 + 1}',
                f'{field1}_str_ten': str(n % 10),
                f'{field1}_emb_{field1}': embdoc1_1
            }),
            cls1(**{
                f'{field1}_int': n * 10 + 2,
                f'{field1}_str': f'str{n * 10 + 2}',
                f'{field1}_str_ten': str(n % 10),
                f'{field1}_emb_{field1}': embdoc1_2
            })
        ]
        doc1_1 = Schema1Doc1(**{f'id': ObjectId(f'{n:024}'),
                                f'doc1_int': n,
                                f'doc1_str': f'str{n}',
                                f'doc1_str_ten': str(n % 10),
                                f'doc1_emblist_{field1}': embdoc1_x}).save()


def get_embedded_jsonpath_parsers(document_type):
    if document_type == '~Schema1EmbDoc1':
        return (
            # doc1.embdoc1
            # doc1.embdoc1.[embdoc2]
            jsonpath_rw.parse(
                'schema1_doc1[*].doc1_emb_embdoc1'
            ),
            # doc1.[embdoc1]
            jsonpath_rw.parse(
                'schema1_doc1[*].doc1_emblist_embdoc1[*]'
            ),
            # doc1.[embdoc1].[embdoc1/embdoc2]
            jsonpath_rw.parse(
                'schema1_doc1[*].doc1_emblist_embdoc1[*].embdoc1_emblist_embdoc1[*]'
            ),
            # doc1.[embdoc1].embdoc1.[embdoc1/embdoc2]
            jsonpath_rw.parse(
                'schema1_doc1[*].doc1_emblist_embdoc1[*].embdoc1_emb_embdoc1[*]'
            ),
            jsonpath_rw.parse(
                'schema1_doc1[*].doc1_emblist_embdoc1[*].embdoc1_emb_embdoc1[*].'
                'embdoc1_emblist_embdoc1[*]'
            ),
        )
    elif document_type == '~Schema1EmbDoc2':
        return (
            # doc1.embdoc1.[embdoc2]
            jsonpath_rw.parse(
                'schema1_doc1[*].doc1_emb_embdoc1.embdoc1_emblist_embdoc2[*]'
            ),
            # doc1.[embdoc1].[embdoc1/embdoc2]
            jsonpath_rw.parse(
                'schema1_doc1[*].doc1_emblist_embdoc1[*].embdoc1_emblist_embdoc2[*]'
            ),
            # doc1.[embdoc1].embdoc1.[embdoc1/embdoc2]
            jsonpath_rw.parse(
                'schema1_doc1[*].doc1_emblist_embdoc1[*].embdoc1_emb_embdoc1[*].'
                'embdoc1_emblist_embdoc2[*]'
            ),
        )
    elif document_type == 'Schema1Doc1':
        return jsonpath_rw.parse('schema1_doc1[*]'),
    else:
        raise ValueError('Unknown document_type')



import pytest

from mongoengine_migrate.actions import RenameDocument
from mongoengine_migrate.exceptions import SchemaError
from mongoengine_migrate.graph import MigrationPolicy


class TestRenameDocument:
    def test_forward__should_do_nothing(self, load_fixture, test_db, dump_db):
        schema = load_fixture('schema1').get_schema()
        dump = dump_db()

        action = RenameDocument('Schema1Doc1', new_name='NewNameDoc')
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert dump == dump_db()

    def test_backward__should_do_nothing(self, load_fixture, test_db, dump_db):
        schema = load_fixture('schema1').get_schema()
        dump = dump_db()

        action = RenameDocument('Schema1Doc1', new_name='NewNameDoc')
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert dump == dump_db()

    def test_prepare__if_such_document_is_not_in_schema__should_raise_error(self,
                                                                            load_fixture,
                                                                            test_db):
        schema = load_fixture('schema1').get_schema()
        del schema['Schema1Doc1']

        action = RenameDocument('Schema1Doc1', new_name='NewNameDoc')

        with pytest.raises(SchemaError):
            action.prepare(test_db, schema, MigrationPolicy.strict)



__all__ = [
    'ByPathContext',
    'ByDocContext',
    'DocumentUpdater',
    'FallbackDocumentUpdater'
]

import logging
from copy import copy
from typing import NamedTuple, Optional, List, Union, Callable, Any, Generator, Tuple

import jsonpath_rw
from pymongo import ReplaceOne
from pymongo.collection import Collection
from pymongo.database import Database
from copy import deepcopy

from mongoengine_migrate import flags
from mongoengine_migrate.graph import MigrationPolicy
from mongoengine_migrate.schema import Schema
from mongoengine_migrate.exceptions import InconsistencyError

log = logging.getLogger('mongoengine-migrate')


def build_array_filters(
        self, value: Optional[Union[Callable, Any]] = None
) -> Optional[List[dict]]:
    """
    Return array filters dict filled out with given value
    :param value: Optional value or callable with one argument
     (array filter value) to substitute to each array filter.
     Default is `{'$exists': True}`
    :return:
    """
    if self.array_filters is None:
        return None

    res = []
    for afilter in self.array_filters:
        if value is None:
            res.append({afilter: {'$exists': True}})
        else:
            res.append({afilter: value(afilter) if callable(value) else value})

    return res


class ByPathContext(NamedTuple):
    """
    Context of `by_path` callback

    Fields:

    * `collection` -- current collection object
    * `filter_dotpath` -- string dotpath to a field or document which
      have to be updated (depends on whether `Updater.field_name` is
      set or not). This field contains only field names separated by
      dots which could be used for filtering in mongodb queries. Could
      be empty string which points to all documents in collection
    * `update_dotpath` -- string dotpath to a field or document which
      have to be updated (depends on whether `Updater.field_name` is
      set or not). This field contains field names with array
      expressions `$[]` separated by dots which could be used in
      update mongodb operations. Could be empty string which points
      to all documents in collection.
    * `array_filters` -- list of array filters used in mongodb
      update operations. `update_docpath` expression relates to this
      field by array filter names. To get value to substitute in an
      update operation please use `build_array_filters` method
    * `extra_filter` -- filter dict which have to AND'ed with
      particular filters in a callback. Typically used for inherited
      documents search.
    """
    collection: Collection
    filter_dotpath: str
    update_dotpath: str
    array_filters: Optional[List[str]]
    extra_filter: dict

    # For IDE method highlight because of workaround below
    build_array_filters = build_array_filters


# Workaround to define methods in NamedTuple for Python 3.6
ByPathContext.build_array_filters = build_array_filters


class ByDocContext(NamedTuple):
    """
    Context of `by_document` callback

    Fields:

    * `collection` -- current collection object
    * `document` -- raw field value which should contain a document.
      Typically this is a dict. But keep in mind that any other value
      could be contained there on some data inconsistency
    * `filter_dotpath` -- string dotpath to a field or document which
      have to be updated (depends on whether `Updater.field_name` is
      set or not). Could be empty string which points to all documents
      in collection
    """
    collection: Optional[Collection]
    document: dict
    filter_dotpath: str


class DocumentUpdater:
    """Document updater class. Used to update certain field in
    collection or embedded document
    """
    def __init__(self,
                 db: Database,
                 document_type: str,
                 db_schema: Schema,
                 field_name: str,
                 migration_policy: MigrationPolicy,
                 document_cls: Optional[str] = None):
        """
        :param db: pymongo database object
        :param document_type: document name
        :param field_name: field to work with
        :param db_schema: current db schema
        :param field_name: If given then update only those records
         which have this field by every dotpath, otherwise update all
         by dotpath. Can be empty string if updater should update
         whole document
        :param migration_policy:
        :param document_cls: if given then we ignore those documents
         and embedded documents whose '_cls' field is not equal to this
         parameter value. Documents with no '_cls' field and fields
         with types other than object will not be ignored. This
         parameter uses for Document inheritance support
        """
        self.db = db
        self.document_type = document_type
        self.db_schema = db_schema
        self.field_name = field_name
        self.migration_policy = migration_policy
        self.document_cls = document_cls
        self._include_missed_fields = False

    @property
    def document_type(self):
        return self._document_type

    @document_type.setter
    def document_type(self, val):
        self._document_type = val
        self.is_embedded = self.document_type.startswith(flags.EMBEDDED_DOCUMENT_NAME_PREFIX)

    def with_missed_fields(self) -> 'DocumentUpdater':
        """Return copy of current Updater which affects all documents
        and embedded documents even if they have missed a field
        """
        res = copy(self)
        res._include_missed_fields = True

        return res

    def update_by_path(self, callback: Callable) -> None:
        """
        Call the given callback for every path to a field contained
        document with needed type.

        For a field in collection it will be called once.

        The same embedded document could be nested or be included to
        many collections -- the callback will be called for each of
        these fields. Returned value is ignored

        Callback parameters are:
        * collection -- pymongo collection object
        * filter_expr -- string dotpath of field which is used in
          filter expressions
        * update_expr -- string dotpath of field (with $[]) which
          is used for pointing
        * array_filters - dict with array filters which could be
          passed to update_many method for instance
        :param callback:
        :return:
        """
        if not self.is_embedded:
            collection_name = self.db_schema[self.document_type].parameters['collection']
            collection = self.db[collection_name]
            self._update_by_path(callback, collection, [], [])
            return

        for collection, update_path, filter_path in self._get_embedded_paths():
            self._update_by_path(callback, collection, filter_path, update_path)

    def update_by_document(self, callback: Callable) -> None:
        """
        Call the given callback for every document of needed
        type found in db. If field contains array of documents then
        callback will be called for each of them.

        Callback function could modify a next field value in-place.
        Returned value is ignored

        Callback parameters are:
        * collection -- pymongo Collection object
        * embedded_document -- field contents, which should contain
          embedded document
        * filter_path -- dotpath of field
        :param callback:
        :return:
        """
        if not self.is_embedded:
            collection_name = self.db_schema[self.document_type].parameters['collection']
            collection = self.db[collection_name]
            self._update_by_document(callback, collection, [], [])
            return

        for collection, update_path, filter_path in self._get_embedded_paths():
            self._update_by_document(callback, collection, filter_path, update_path)

    def update_combined(self,
                        by_path_cb: Callable,
                        by_doc_cb: Callable,
                        embedded_nonarray_by_doc: bool = True,
                        embedded_array_by_doc: bool = True) -> None:
        """
        Perform an update depending on document type. Usual documents
        will get updated using by_path.
        If `embedded_nonarray_by_doc` is True then embedded documents
        which paths do not contain arrays (with "$[]") will be
        updated using by_doc, otherwise using by_path.
        If `embedded_array_by_doc` is True then embedded documents
        which paths contain arrays (with "$[]") will be updated using
        by_doc, otherwise using by_path.

        Some MongoDB queries/pipelines are unable to be executed over
        embedded field dotpath. For example, `$rename` is unsuitable
        for array updates of embedded documents (path contains
        "$[]"), but works for non-array updates. Some queries
        (`$set` for example) accept both variants. In the same time
        MongoDB queries work with document field without restrictions.

        In order to manage with this restrictions we could update
        documents by hand in python if the query is limited to do this.

        :param by_path_cb: "by_path" callback which is always
         called for non-embedded document
        :param by_doc_cb: "by_document" callback
         which is called always for non-array dotpaths.
        :param embedded_array_by_doc: if True then embedded docs with
         array dotpaths (with "$[]") will get updated using
         by_doc callback, or by_path otherwise.
        :param embedded_nonarray_by_doc: if True then embedded docs with
         non-array dotpaths (without "$[]") will get updated using
         by_doc callback, or by_path otherwise.
        """
        if self.is_embedded:
            for collection, update_path, filter_path in self._get_embedded_paths():
                is_array_update = bool('$[]' in update_path)
                call_by_doc = is_array_update and embedded_array_by_doc \
                    or not is_array_update and embedded_nonarray_by_doc

                if call_by_doc:
                    self._update_by_document(by_doc_cb, collection, filter_path, update_path)
                else:
                    self._update_by_path(by_path_cb, collection, filter_path, update_path)
        else:
            collection_name = self.db_schema[self.document_type].parameters['collection']
            collection = self.db[collection_name]
            self._update_by_path(by_path_cb, collection, [], [])

    def _update_by_path(self,
                        callback: Callable,
                        collection: Collection,
                        filter_path: List[str],
                        update_path: List[str]) -> None:
        if self.field_name:
            filter_path = filter_path + [self.field_name]  # Don't modify filter_path
            update_path = update_path + [self.field_name]  #

        update_path, array_filters = self._inject_array_filters(update_path)
        extra_filter = {'_cls': self.document_cls} if self.document_cls else {}

        filter_dotpath = '.'.join(filter_path)
        update_dotpath = '.'.join(update_path)
        ctx = ByPathContext(collection=collection,
                            filter_dotpath=filter_dotpath,
                            update_dotpath=update_dotpath,
                            array_filters=array_filters,
                            extra_filter=extra_filter)
        callback(ctx)

    def _update_by_document(self,
                            callback: Callable,
                            collection: Collection,
                            filter_path: List[str],
                            update_path: List[str]) -> None:
        """
        Call a callback for every document found by given filterpath
        :param callback: by_doc callback
        :param collection: pymongo.Collection object
        :param filter_path: filter dotpath to substitute to find()
        :param update_path: Update dotpath (with $[]) is
         pointed which document to pick and call the callback
         for each of them (nested array of embedded documents for
         instance). If None is passed then we pick a document itself
        :return:
        """
        field_filter_path = copy(filter_path)
        if self.field_name:
            field_filter_path += [self.field_name]
        filter_dotpath = '.'.join(field_filter_path)

        if not update_path:
            json_path = '$'  # update_path points to any document
        else:
            # update_path is mongo update path
            json_path = '.'.join(f.replace('$[]', '[*]') for f in update_path)
            json_path = json_path.replace('.[*]', '[*]')
        parser = jsonpath_rw.parse(json_path)

        find_fltr = {}
        if not self._include_missed_fields and filter_dotpath:
            find_fltr = {filter_dotpath: {'$exists': True}}
        if self.document_cls:
            find_fltr['_cls'] = self.document_cls

        if flags.dry_run:
            msg = '* db.%s.find(%s) -> [Loop](%s) -> db.%s.bulk_write(...)'
            log.info(msg, collection.name, find_fltr, filter_dotpath, collection.name)
            return

        bulk_db = flags.database2
        bulk_collection = bulk_db[collection.name]

        buf = []
        for doc in collection.find(find_fltr):
            prev_doc = deepcopy(doc)

            # Recursively apply the callback to every embedded doc
            for embedded_doc in parser.find(doc):
                embedded_doc = embedded_doc.value
                if self.document_cls:
                    if embedded_doc is None:
                        continue
                    if not isinstance(embedded_doc, dict):
                        # Field contains smth another than embedded doc
                        if self.migration_policy.name == 'strict':
                            raise InconsistencyError(
                                f"Field {filter_dotpath} has wrong value {embedded_doc!r} "
                                f"(should be embedded document) in record {doc}"
                            )
                        else:
                            continue
                    if embedded_doc.get('_cls', self.document_cls) != self.document_cls:
                        # Skip since document doesn't belong to
                        # document class (document inheritance,
                        # DynamicField)
                        # See `DocumentMetaclass` implementation
                        continue
                ctx = ByDocContext(collection=collection,
                                   document=embedded_doc,
                                   filter_dotpath=filter_dotpath)
                # Callback should change a dict in-place
                callback(ctx)

            # Write a document only if it was changed by callback
            if prev_doc != doc:
                buf.append(ReplaceOne({'_id': doc['_id']}, doc, upsert=False))

            # Flush buffer
            if len(buf) >= flags.BULK_BUFFER_LENGTH:
                bulk_collection.bulk_write(buf, ordered=False)
                buf.clear()
        if buf:
            bulk_collection.bulk_write(buf, ordered=False)
            buf.clear()

    def _get_embedded_paths(self) -> Generator[Tuple[Collection, list, list], None, None]:
        """
        Return dotpaths to fields of embedded documents found in db and
        collection object where they was found.
        Returned dotpaths are filter expressions (pure dotpath)
        and update expressions (dotpath with `$[]` expressions)
        :return: tuple(collection_object, update_dotpath, filter_dotpath)
        """
        # Current document_type is not EmbeddedDocument, so not
        # embedded paths can be produced
        if not self.is_embedded:
            return

        # Document types of non-embedded documents
        document_types = ((name, schema) for name, schema in self.db_schema.items()
                          if not name.startswith(flags.EMBEDDED_DOCUMENT_NAME_PREFIX))

        for document_type, document_schema in document_types:
            collection = self.db[document_schema.parameters['collection']]
            for path in self._find_embedded_fields(collection,
                                                   document_type,
                                                   self.document_type,
                                                   self.db_schema):
                update_path = path  # type: list
                filter_path = [p for p in path if p != '$[]']

                yield collection, update_path, filter_path

    def _find_embedded_fields(self,
                              collection: Collection,
                              root_doctype: str,
                              search_doctype: str,
                              db_schema: Schema,
                              _base_path: Optional[list] = None) -> Generator[list, None, None]:
        """
        Perform recursive search for embedded document fields of given
        type in given collection and return key paths to them. Paths
        for fields which are contained objects and arrays in database
        are returned separately: ['a', 'b', 'c'] and
        ['a', 'b', '$[]', 'c', '$[]'] (b and c are arrays) appropriately

        Each key path is returned if it actually exists in db. This
        check is needed to break recursion since embedded documents
        may refer to each other or even themselves.
        :param collection: collection object where to search given
         embedded document
        :param root_doctype: document type name where to perform
         recursive search
        :param search_doctype: embedded document name to search
        :param db_schema: db schema
        :return:
        """
        if _base_path is None:
            _base_path = []

        # Restrict recursion depth
        max_path_len = 64
        if len(_base_path) >= max_path_len:
            return

        # Return every field nested path if it has a needed type_key.
        # Next also overlook in depth to each embedded document field
        # (including fields with another type_keys) if they have nested
        # embedded documents which we also should to check. Recursion
        # stops when we found that nested field which does not exist
        # in db. Keep in mind that embedded documents could refer to
        # each other or even to itself.
        #
        # Fields may contain embedded docs and/or array of embedded docs
        # Because of limitations of MongoDB we're checking type
        # (object/array) and update a field further separately.
        for field, field_schema in db_schema.get(root_doctype, {}).items():
            path = _base_path + [field]
            filter_path = [p for p in path if p != '$[]']

            # Check if field is EmbeddedField or EmbeddedFieldList
            ref = field_schema.get('target_doctype')
            if ref is None or not ref.startswith(flags.EMBEDDED_DOCUMENT_NAME_PREFIX):
                # Skip fields which don't point to embedded document
                continue

            # Check if field type is object or array.
            # Dotpath field resolving always takes the first
            # element type if it is an array
            # So do the {"field.path.0": {$exists: true}} in
            # order to ensure that field contains array (non-empty)
            array_dotpath = '.'.join(filter_path + ['0'])

            object_results = collection.count_documents(
                {
                    array_dotpath: {'$exists': False},
                    '.'.join(filter_path): {'$type': "object"}
                },
                limit=1
            )
            if object_results > 0:
                if ref == search_doctype:
                    yield path
                yield from self._find_embedded_fields(collection,
                                                      ref,
                                                      search_doctype,
                                                      db_schema,
                                                      path)
                # Return if field contains objects
                # It's better to have ability to handle situation
                # when the same field has both array and object
                # values at the same time.
                # But this function tests field existence using
                # dotpath. Dotpath resolving makes no distinction
                # between array and object, so it can be generated
                # extra paths.
                # For example, a field could contain array which
                # contains objects only and also could contain
                # object which contains arrays only. Function must
                # return two paths: object->arrays, array->objects.
                # But because of dotpath resolving thing we'll
                # got all 4 path combinations.
                # For a while I leave return here. It's better to
                # remove it and solve the problem somehow.
                # TODO: I'll be back
                continue

            # TODO: return also empty array fields
            array_results = collection.count_documents(
                {array_dotpath: {'$exists': True}},
                limit=1
            )
            if array_results > 0:
                if ref == search_doctype:
                    yield path + ['$[]']
                yield from self._find_embedded_fields(collection,
                                                      ref,
                                                      search_doctype,
                                                      db_schema,
                                                      path + ['$[]'])

    def _inject_array_filters(self, update_path: list) -> Tuple[list, Optional[list]]:
        """
        Inject array filters for each array field path.
        Return modified update path and appropriate array filters
        :param update_path:
        :return: tuple(update_path, array_filters)
        """
        #
        array_filters = {}
        update_path = update_path.copy()
        for num, item in enumerate(update_path):
            if item == '$[]':
                update_path[num] = f'$[elem{num}]'
                array_filters[f'elem{num}.{update_path[num + 1]}'] = None

        return update_path, (list(array_filters.keys()) or None)


class FallbackDocumentUpdater(DocumentUpdater):
    """Fallback variant of DocumentUpdater. All update operations here
    will be performed using by by_doc callback
    """
    def __init__(self, updater: DocumentUpdater):
        """Copying constructor"""
        super().__init__(updater.db, updater.document_type, updater.db_schema,
                         updater.field_name, updater.migration_policy, updater.document_cls)
        self._include_missed_fields = updater._include_missed_fields

    def update_combined(self,
                        by_path_cb: Callable,
                        by_doc_cb: Callable,
                        embedded_nonarray_by_doc: bool = True,
                        embedded_array_by_doc: bool = True) -> None:
        """
        Update document using by_doc callback only. Fallback variant
        of the same function in DocumentUpdater
        :param by_path_cb: not used
        :param by_doc_cb: by_doc callback
        :param embedded_nonarray_by_doc: not used
        :param embedded_array_by_doc: not used
        :return:
        """
        self.update_by_document(by_doc_cb)

    def _update_by_path(self,
                        callback: Callable,
                        collection: Collection,
                        filter_path: List[str],
                        update_path: List[str],
                        extra_filter: Optional[dict] = None) -> None:
        raise AttributeError('This version of MongoDB does not support such query')



__all__ = [
    'FieldHandlerMeta',
    'CommonFieldHandler',
]

import inspect
import weakref
from typing import Type, Iterable, List, Tuple, Collection, Any

import mongoengine.fields
from pymongo.database import Database

import mongoengine_migrate.flags as flags
from mongoengine_migrate.exceptions import SchemaError, MigrationError, ActionError
from mongoengine_migrate.fields.registry import type_key_registry, add_field_handler
from mongoengine_migrate.graph import MigrationPolicy
from mongoengine_migrate.mongo import check_empty_result
from mongoengine_migrate.schema import Schema
from mongoengine_migrate.utils import get_closest_parent, document_type_to_class_name, Diff, UNSET
from .registry import CONVERTION_MATRIX
from ..updater import ByPathContext, ByDocContext, DocumentUpdater


class FieldHandlerMeta(type):
    def __new__(mcs, name, bases, attrs):
        me_classes_attr = 'field_classes'
        me_classes = attrs.get(me_classes_attr)

        assert not {'field_name', 'document_type'} & attrs.get('schema_skel_keys', set()), \
            "Handler schema_skel_keys shouldn't have keys matched with BaseAction parameters"

        assert isinstance(me_classes, (List, Tuple)) or me_classes is None, \
            f'{me_classes_attr} must be mongoengine field classes list'

        attrs['_meta'] = weakref.proxy(mcs)

        klass = super(FieldHandlerMeta, mcs).__new__(mcs, name, bases, attrs)
        if me_classes:
            for me_class in me_classes:
                add_field_handler(me_class, klass)

        return klass


class CommonFieldHandler(metaclass=FieldHandlerMeta):
    """
    Default handler for all mongoengine fields. Used for fields which
    does not have specific handler.

    Also this is base class for other field-specific handlers
    """
    #: Mongoengine field classes which concrete handler can be used for
    field_classes: Iterable[Type[mongoengine.fields.BaseField]] = [
        mongoengine.fields.BaseField
    ]

    #: Schema keys used for all fields listed in `field_classes`
    #: This variable inherits, i.e. keys defined in current class is
    #: appended to keys defined in parent classes
    schema_skel_keys: Iterable[str] = {
        'db_field', 'required', 'default', 'unique', 'unique_with', 'primary_key', 'choices',
        'null', 'sparse', 'type_key'
    }

    def __init__(self,
                 db: Database,
                 document_type: str,
                 left_schema: Schema,
                 left_field_schema: dict,  # FIXME: get rid of
                 right_field_schema: dict,
                 migration_policy: MigrationPolicy):
        """
        :param db: pymongo Database object
        :param document_type: collection name. Could contain
         collection name or embedded document name
        :param left_schema:
        :param left_field_schema:
        :param right_field_schema:
        :param migration_policy:
        """
        self.db = db
        self.document_type = document_type
        self.left_schema = left_schema
        self.left_field_schema = left_field_schema
        self.right_field_schema = right_field_schema
        self.migration_policy = migration_policy

        self.is_embedded = self.document_type.startswith(flags.EMBEDDED_DOCUMENT_NAME_PREFIX)
        self.collection = None
        if not self.is_embedded:
            collection_name = left_schema[document_type].parameters['collection']
            self.collection = None if self.is_embedded else db[collection_name]

    @classmethod
    def schema_skel(cls) -> dict:
        """
        Return db schema skeleton dict, which contains keys taken from
        `schema_skel_keys` and Nones as values
        """
        keys = []
        for klass in reversed(inspect.getmro(cls)):
            keys.extend(getattr(klass, 'schema_skel_keys', []))

        return {f: None for f in keys}

    @classmethod
    def build_schema(cls, field_obj: mongoengine.fields.BaseField) -> dict:
        """
        Return schema of a given mongoengine field object

        'type_key' schema item will get filled with a mongoengine field
        class name or one of its parents which have its own type key
        in registry
        :param field_obj: mongoengine field object
        :return: field schema dict
        """
        schema_skel = cls.schema_skel()
        schema = {f: getattr(field_obj, f, val) for f, val in schema_skel.items()}

        if 'default' in schema:
            schema['default'] = cls._normalize_default(schema['default'])

        if 'choices' in schema:
            schema['choices'] = cls._normalize_choices(schema['choices'])

        field_class = field_obj.__class__
        if field_class.__name__ in type_key_registry:
            schema['type_key'] = field_class.__name__
        else:
            registry_field_cls = get_closest_parent(
                field_class,
                (x.field_cls for x in type_key_registry.values())
            )
            if registry_field_cls is None:
                raise ActionError(f'Could not find {field_class!r} or one of its base classes '
                                  f'in type_key registry')

            schema['type_key'] = registry_field_cls.__name__

        return schema

    def change_param(self, db_field: str, name: str):
        """
        DB commands to be run in order to change given parameter

        This is a facade method which calls concrete method which
        changes given parameter. Such methods should have name
        'change_NAME' where NAME is a parameter name.
        :param db_field: db field name to change
        :param name: parameter name to change
        :return:
        """
        assert name != 'param', "Schema key could not be 'param'"

        diff = Diff(
            old=self.left_field_schema.get(name, UNSET),
            new=self.right_field_schema.get(name, UNSET),
            key=name
        )

        method = getattr(self, f'change_{name}')
        inherit = self.left_schema[self.document_type].parameters.get('inherit')
        document_cls = document_type_to_class_name(self.document_type) if inherit else None
        updater = DocumentUpdater(self.db, self.document_type, self.left_schema, db_field,
                                  self.migration_policy, document_cls)
        return method(updater, diff)

    def change_db_field(self, updater: DocumentUpdater, diff: Diff):
        """
        Change db field name of a field. Simply rename this field
        :param updater:
        :param diff:
        :return:
        """
        def by_path(ctx: ByPathContext):
            path = ctx.filter_dotpath.split('.')[:-1]
            ctx.collection.update_many(
                {ctx.filter_dotpath: {'$exists': True}, **ctx.extra_filter},
                {'$rename': {ctx.filter_dotpath: '.'.join(path + [diff.new])}}
            )

        def by_doc(ctx: ByDocContext):
            doc = ctx.document
            if diff.old in doc:
                doc[diff.new] = doc.pop(diff.old)

        self._check_diff(updater, diff, False, str)
        if not diff.new or not diff.old:
            raise SchemaError(f"{updater.document_type}{updater.field_name}.db_field "
                              f"must be a non-empty string")

        updater.update_combined(by_path, by_doc, False)

    def change_required(self, updater: DocumentUpdater, diff: Diff):
        """
        Make field required, which means to add this field to all
        documents. Reverting of this doesn't require smth to do
        :param updater:
        :param diff:
        :return:
        """
        self._check_diff(updater, diff, False, bool)
        if diff.old is not True and diff.new is True:
            default = self.right_field_schema.get('default')
            # None and UNSET default has the same meaning here
            if default is None:
                raise SchemaError(f'{updater.document_type}{updater.field_name}.default is not '
                                  f'set for required field')

            self._set_default_value(updater, default)

    def change_default(self, updater: DocumentUpdater, diff: Diff):
        """Stub method. No need to do smth on default change"""
        pass

    def change_unique(self, updater: DocumentUpdater, diff: Diff):
        # TODO
        pass

    def change_unique_with(self, updater: DocumentUpdater, diff: Diff):
        # TODO
        pass

    def change_primary_key(self, updater: DocumentUpdater, diff: Diff):
        """
        Setting field as primary key means to set it required and unique
        :param updater:
        :param diff:
        :return:
        """
        def by_path(ctx: ByPathContext):
            fltr = {ctx.filter_dotpath: {'$exists': False}, **ctx.extra_filter}
            check_empty_result(ctx.collection, ctx.filter_dotpath, fltr)

        self._check_diff(updater, diff, False, bool)

        if updater.is_embedded:
            raise SchemaError(f'Embedded document {updater.document_type} cannot have primary key')
        if self.migration_policy.name == 'strict':
            updater.update_by_path(by_path)
        # self.change_unique([], []) or []  # TODO

    # TODO: consider Document, EmbeddedDocument as choices
    def change_choices(self, updater: DocumentUpdater, diff: Diff):
        """
        Set choices for a field
        :param updater:
        :param diff:
        :return:
        """
        def by_path(ctx: ByPathContext):
            choices = diff.new
            fltr = {
                ctx.filter_dotpath: {'$nin': choices, '$exists': True},
                **ctx.extra_filter
            }
            check_empty_result(ctx.collection, ctx.filter_dotpath, fltr)

        self._check_diff(updater, diff, True, Collection)

        if diff.new is not None and self.migration_policy.name == 'strict':
            updater.update_by_path(by_path)

    def change_null(self, updater: DocumentUpdater, diff: Diff):
        """Force set db fields to `null` if they are was not set and
        they defined with `null=True` in mongoengine
        """
        self._check_diff(updater, diff, False, bool)
        if diff.new is True:
            self._set_default_value(updater, None)

    def change_sparse(self, updater: DocumentUpdater, diff: Diff):
        pass

    def change_type_key(self, updater: DocumentUpdater, diff: Diff):
        """
        Change type of field. Try to convert value in db
        :param updater:
        :param diff:
        :return:
        """
        self._check_diff(updater, diff, False, str)
        if not diff.old or not diff.new:
            raise SchemaError(f"Old or new {updater.document_type}{updater.field_name}.type_key "
                              f"values are not set")

        field_classes = []
        for val in (diff.old, diff.new):
            if val not in type_key_registry:
                raise SchemaError(f'Unknown type_key {updater.document_type}{updater.field_name}: '
                                  f'{val!r}')
            field_classes.append(type_key_registry[val].field_cls)

        new_handler_cls = type_key_registry[diff.new].field_handler_cls
        new_handler = new_handler_cls(self.db,
                                      self.document_type,
                                      self.left_schema,
                                      self.left_field_schema,
                                      self.right_field_schema,
                                      self.migration_policy)
        new_handler.convert_type(updater, *field_classes)

    def convert_type(self,
                     updater: DocumentUpdater,
                     from_field_cls: Type[mongoengine.fields.BaseField],
                     to_field_cls: Type[mongoengine.fields.BaseField]):
        """
        Convert field type from another to a current one. This method
        is called only if such change was requested in a migration.

        We use convertion matrix here. It contains mapping between
        old and new types and appropriate converter function which
        is called to perform such convertion.

        Old field can be either actual field type which used before.
        But in case if that field was a user defined class and does
        not exist already, the BaseField will be sent.

        New field will always have target mongoengine field type
        :param updater:
        :param from_field_cls: mongoengine field class which was used
         before
        :param to_field_cls: mongoengine field class which will be used
         further
        :return:
        """

        type_converters = CONVERTION_MATRIX.get(from_field_cls) or \
            CONVERTION_MATRIX.get(get_closest_parent(from_field_cls, CONVERTION_MATRIX.keys()))

        if type_converters is None:
            raise MigrationError(f'Type converter not found for convertion '
                                 f'{from_field_cls!r} -> {to_field_cls!r}')

        type_converter = type_converters.get(to_field_cls) or \
            type_converters.get(get_closest_parent(to_field_cls, type_converters))

        if type_converter is None:
            raise MigrationError(f'Type converter not found for convertion '
                                 f'{from_field_cls!r} -> {to_field_cls!r}')

        type_converter(updater)

    @staticmethod
    def _check_diff(updater: DocumentUpdater, diff: Diff, can_be_none=True, check_type=None):
        if diff.new == diff.old:
            raise SchemaError(f'{updater.document_type}.{updater.field_name}.{diff.key} '
                              f'does not changed from previous Action')

        if check_type is not None:
            if diff.old not in (UNSET, None) and not isinstance(diff.old, check_type) \
                    or diff.new not in (UNSET, None) and not isinstance(diff.new, check_type):
                raise SchemaError(f'{updater.document_type}.{updater.field_name}.{diff.key} '
                                  f'must have type {check_type!r}')

        if not can_be_none:
            if diff.old is None or diff.new is None:
                raise SchemaError(f'{updater.document_type}.{updater.field_name}.{diff.key} '
                                  f'could not be None')

    @classmethod
    def _normalize_default(cls, default):
        if callable(default):
            default = default()

        # Check if the expression repr produces correct python expr
        try:
            eval(repr(default), {}, {})
        except:
            # FIXME: with required=True this leads to error (mongoengine will not raise error since default is specified)
            default = None

        return default

    @classmethod
    def _normalize_choices(cls, choices):
        if not isinstance(choices, Iterable):
            return None

        if isinstance(next(iter(choices)), (list, tuple)):
            # next(iter) is useful for sets
            choices = [k for k, _ in choices]
        else:
            # Make a tuple from any iterable type (e.g. dict_keys)
            choices = tuple(choices)

        return choices

    @staticmethod
    def _set_default_value(updater: DocumentUpdater, value: Any):
        """Set a given value only for unset fields"""
        def by_path(ctx: ByPathContext):
            # Update documents only
            ctx.collection.update_many(
                {ctx.filter_dotpath: {'$exists': False}, **ctx.extra_filter},
                {'$set': {ctx.update_dotpath: value}},
                array_filters=ctx.build_array_filters()
            )

        def by_doc(ctx: ByDocContext):
            # Update embedded documents
            if updater.field_name not in ctx.document:
                ctx.document.setdefault(updater.field_name, value)

        updater.with_missed_fields().update_combined(by_path, by_doc)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='mongoengine-migrate',
    description='Migrations for MongoEngine ODM inspired by Django',
    version='0.0.1a3',  # Also pyproject.toml
    author='Igor Derkach',
    author_email='gosha753951@gmail.com',
    url='https://github.com/bdragon300/mongoengine-migrate',
    license='Apache-2.0',
    python_requires='>=3.6',
    packages=setuptools.find_packages(exclude=['tests']),
    package_data={'mongoengine_migrate': ['migration_template.tpl']},
    long_description=long_description,
    long_description_content_type='text/markdown',
    entry_points={"console_scripts": ["mongoengine_migrate=mongoengine_migrate.cli:cli"]},
    keywords=["mongo", "mongodb", "mongoengine", "migrate", "migration"],
    classifiers=[
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: Apache Software License',
        'Intended Audience :: Developers',
        'Topic :: Database'
    ],
    # Also tox.ini
    install_requires=[
        'mongoengine>=0.16.0',
        'pymongo>=3.0',
        'dictdiffer>=0.7.0',
        'jinja2',
        'click',
        'wrapt',
        'python-dateutil',
        'jsonpath_rw'
    ],
)



__all__ = [
    'MigrationPolicy',
    'Migration',
    'MigrationsGraph'
]

from typing import Dict, List

from mongoengine_migrate.exceptions import MigrationGraphError
from mongoengine_migrate.utils import Slotinit
from enum import Enum


class MigrationPolicy(Enum):
    """Policy which determines how much the migration engine will
    be tolerant to already existing data which does not satisfy to
    mongoengine schema.

    MongoDB allows to keep any type of data in any field because of
    its schemaless nature. The data could be fed from various sources,
    not only those one with our mongoengine documents, MongoDB does not
    check their type.

    When `strict` policy is used the presence of data which does not
    satisfy to schema is treated as error and migration process will
    be interrupted.

    Using `relaxed` policy the migration engine does not check
    existing value types and tries to migrate data which related only to
    migration itself. E.g. when we change a field type from `long` to
    `string`, only `long` values will be casted to string type, but
    existing objectId or embedded documents or smth else in the same
    field will be skipped.
    """
    strict = 0
    relaxed = 1


class Migration(Slotinit):
    """Object represents one migration

    Contains information which is set in migration:
    * name -- migration file name without '.py' suffix
    * dependencies -- name list of migrations which this migration is
      dependent by
    * applied -- is migration was applied or not. Taken from database
    """
    __slots__ = ('name', 'dependencies', 'applied', 'module')
    defaults = {'applied': False}

    def get_actions(self):
        # FIXME: type checking, attribute checking
        # FIXME: tests
        return self.module.actions

    @property
    def policy(self) -> MigrationPolicy:
        attr = getattr(self.module, 'policy', MigrationPolicy.strict.name)
        return getattr(MigrationPolicy, attr)


class MigrationsGraph:
    # TODO: make it dict-like, not list-like
    def __init__(self):
        # Following two variables contains the same migrations digraph
        # but from different points of view
        self._parents: Dict[str, List[Migration]] = {}  # {child_name: [parent_obj...]}
        self._children: Dict[str, List[Migration]] = {}  # {parent_name: [child_obj...]}

        self._migrations: Dict[str, Migration] = {}  # {migration_name: migration_obj}

    @property
    def initial(self):
        """Return initial migration object"""
        for name, parents in self._parents.items():
            if not parents:
                return self._migrations[name]

    @property
    def last(self):
        """Return last children migration object"""
        for name, children in self._children.items():
            if not children:
                return self._migrations[name]

    @property
    def migrations(self):
        """Migration objects dict"""
        return self._migrations

    def add(self, migration: Migration):
        """
        Add migration to the graph. If object with that name exists
        in graph then it will be replaced.
        :param migration: Migration object
        :return:
        """
        self._parents[migration.name] = []
        self._children[migration.name] = []

        for partner in self._migrations.values():
            if partner.name == migration.name:
                continue
            if partner.name in migration.dependencies:
                self._parents[migration.name].append(partner)
                self._children[partner.name].append(migration)
            if migration.name in partner.dependencies:
                self._children[migration.name].append(partner)
                self._parents[partner.name].append(migration)

        self._migrations[migration.name] = migration

    def clear(self):  # TODO: tests
        """
        Clear graph
        :return:
        """
        self._parents = {}
        self._children = {}
        self._migrations = {}

    def verify(self):
        """
        Verify migrations graph to be satisfied to consistency rules
        Graph must not have loops, disconnections.
        Also it should have single initial migration and (for a while)
        single last migration.
        :raises MigrationGraphError: if problem in graph was found
        :return:
        """
        # FIXME: This function is not used anywhere
        initials = []
        last_children = []

        for name, obj in self._migrations.items():
            if not self._parents[name]:
                initials.append(name)
            if not self._children[name]:
                last_children.append(name)
            if len(obj.dependencies) > len(self._parents[name]):
                diff = set(obj.dependencies) - {x.name for x in self._parents[name]}
                raise MigrationGraphError(f'Unknown dependencies in migration {name!r}: {diff}')
            if name in (x.name for x in self._children[name]):
                raise MigrationGraphError(f'Found migration which dependent on itself: {name!r}')

        if len(initials) == len(last_children) and len(initials) > 1:
            raise MigrationGraphError(f'Migrations graph is disconnected, history segments '
                                 f'started on: {initials!r}, ended on: {last_children!r}')
        if len(initials) > 1:
            raise MigrationGraphError(f'Several initial migrations found: {initials!r}')

        if len(last_children) > 1:
            raise MigrationGraphError(f'Several last migrations found: {last_children!r}')

        if not initials or not last_children:
            raise MigrationGraphError(f'No initial or last children found')

    def walk_down(self, from_node: Migration, unapplied_only=True, _node_counters=None):
        """
        Walks down over migrations graph. Iterates in order as migrations
        should be applied.

        We're used modified DFS (depth-first search) algorithm to traverse
        the graph. Migrations are built into directed graph (digraph)
        counted from one root node to the last ones. Commonly DFS tries to
        walk to the maximum depth firstly.

        But there are some problems:
        * Graph can have directed cycles. This happens when some
          migration has several dependencies. Therefore we'll walk
          over such migration several times
        * Another problem arises from the first one. Typically we must walk
          over all dependencies before a dependent migration will be
          touched. DFS will process only one dependency before get to
          a dependent migration

        In order to manage it we use counter for each migration
        (node in digraph) initially equal to its parents count.
        Every time the algorithm gets to node it decrements this counter.
        If counter > 0 after that then don't touch this node and
        break traversing on this depth and go up. If counter == 0 then
        continue traversing.
        :param from_node: current node in graph
        :param unapplied_only: if True then return only unapplied migrations
         or return all migrations otherwise
        :param _node_counters:
        :raises MigrationGraphError: if graph has a closed cycle
        :return: Migration objects generator
        """
        # FIXME: may yield nodes not related to target migration if branchy graph
        # FIXME: if migration was applied after its dependencies unapplied then it is an error
        # FIXME: should have stable migrations order
        if _node_counters is None:
            _node_counters = {}
        if from_node is None:
            return ()
        _node_counters.setdefault(from_node.name, len(self._parents[from_node.name]) or 1)
        _node_counters[from_node.name] -= 1

        if _node_counters[from_node.name] > 0:
            # Stop on this depth if not all parents has been viewed
            return

        if _node_counters[from_node.name] < 0:
            # A node was already returned and we're reached it again
            # This means there is a closed cycle
            raise MigrationGraphError(f'Found closed cycle in migration graph, '
                                      f'{from_node.name!r} is repeated twice')

        if not (from_node.applied and unapplied_only):
            yield from_node

        for child in self._children[from_node.name]:
            yield from self.walk_down(child, unapplied_only, _node_counters)

    def walk_up(self, from_node: Migration, applied_only=True, _node_counters=None):
        """
        Walks up over migrations graph. Iterates in order as migrations
        should be reverted.

        We're using modified DFS (depth-first search) algorithm which in
        reversed order (see `walk_down`). Instead of looking at node
        parents count we're consider children count in order to return
        all dependent nodes before dependency.

        Because of the migrations graph may have many orphan child nodes
        they all should be passed as parameter
        :param from_node:  last children node we are starting for
        :param applied_only: if True then return only applied migrations,
         return all migrations otherwise
        :param _node_counters:
        :raises MigrationGraphError: if graph has a closed cycle
        :return: Migration objects generator
        """
        # FIXME: may yield nodes not related to reverting if branchy graph
        # FIXME: if migration was unapplied before its dependencies applied then it is an error
        if _node_counters is None:
            _node_counters = {}
        if from_node is None:
            return ()
        _node_counters.setdefault(from_node.name, len(self._children[from_node.name]) or 1)
        _node_counters[from_node.name] -= 1

        if _node_counters[from_node.name] > 0:
            # Stop in this depth if not all children has been viewed
            return

        if _node_counters[from_node.name] < 0:
            # A node was already returned and we're reached it again
            # This means there is a closed cycle
            raise MigrationGraphError(f'Found closed cycle in migration graph, '
                                      f'{from_node.name!r} is repeated twice')

        if from_node.applied or not applied_only:
            yield from_node

        for child in self._parents[from_node.name]:
            yield from self.walk_up(child, applied_only, _node_counters)

    def __iter__(self):
        return iter(self.walk_down(self.initial, unapplied_only=False))

    def __reversed__(self):
        return iter(self.walk_up(self.last, applied_only=False))

    def __contains__(self, migration: Migration):
        return migration in self._migrations.values()

    def __eq__(self, other):
        if other is self:
            return True
        if not isinstance(other, MigrationsGraph):
            return False

        return all(migr == other._migrations.get(name) for name, migr in self._migrations.items())

    def __ne__(self, other):
        return not self.__eq__(other)
